import pytest

import pyteal as pt
from pyteal.ast.txn import Txn, TxnField, TxnType
from pyteal.types import types_match

avm4Options = pt.CompileOptions(version=4)
avm5Options = pt.CompileOptions(version=5)
avm6Options = pt.CompileOptions(version=6)


def test_InnerTxnBuilder_Begin():
    expr = pt.InnerTxnBuilder.Begin()
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.itxn_begin)])

    actual, _ = expr.__teal__(avm5Options)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm4Options)


def test_InnerTxnBuilder_Submit():
    expr = pt.InnerTxnBuilder.Submit()
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.itxn_submit)])

    actual, _ = expr.__teal__(avm5Options)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm4Options)


def test_InnerTxnBuilder_Next():
    expr = pt.InnerTxnBuilder.Next()
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.itxn_next)])

    actual, _ = expr.__teal__(avm6Options)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm5Options)


def test_InnerTxnBuilder_SetField():
    for field in pt.TxnField:
        if field.is_array:
            with pytest.raises(pt.TealInputError):
                pt.InnerTxnBuilder.SetField(field, pt.Int(0))
            continue

        for value, opArgs in (
            (pt.Int(0), (pt.Op.int, 0)),
            (pt.Bytes("value"), (pt.Op.byte, '"value"')),
        ):
            assert field.type_of() in (pt.TealType.uint64, pt.TealType.bytes)

            if not types_match(field.type_of(), value.type_of()):
                with pytest.raises(pt.TealTypeError):
                    pt.InnerTxnBuilder.SetField(field, value)
                continue

            expr = pt.InnerTxnBuilder.SetField(field, value)
            assert expr.type_of() == pt.TealType.none
            assert not expr.has_return()

            expected = pt.TealSimpleBlock(
                [
                    pt.TealOp(value, *opArgs),
                    pt.TealOp(expr, pt.Op.itxn_field, field.arg_name),
                ]
            )

            actual, _ = expr.__teal__(avm5Options)
            actual.addIncoming()
            actual = pt.TealBlock.NormalizeBlocks(actual)

            assert actual == expected

            with pytest.raises(pt.TealInputError):
                expr.__teal__(avm4Options)


ITXN_FIELDS_CASES = [
    ({}, pt.Seq()),
    (
        {pt.TxnField.amount: pt.Int(5)},
        pt.InnerTxnBuilder.SetField(pt.TxnField.amount, pt.Int(5)),
    ),
    (
        {
            pt.TxnField.type_enum: pt.TxnType.Payment,
            pt.TxnField.close_remainder_to: pt.Txn.sender(),
        },
        pt.Seq(
            pt.InnerTxnBuilder.SetField(pt.TxnField.type_enum, pt.TxnType.Payment),
            pt.InnerTxnBuilder.SetField(
                pt.TxnField.close_remainder_to, pt.Txn.sender()
            ),
        ),
    ),
    (
        {pt.TxnField.accounts: pt.Txn.accounts},
        pt.For(
            (i := pt.ScratchVar()).store(pt.Int(0)),
            i.load() < pt.Txn.accounts.length(),
            i.store(i.load() + pt.Int(1)),
        ).Do(
            pt.InnerTxnBuilder.SetField(pt.TxnField.accounts, [Txn.accounts[i.load()]])
        ),
    ),
]


def test_InnerTxnBuilder_SetFields():
    for fields, expectedExpr in ITXN_FIELDS_CASES:
        expr = pt.InnerTxnBuilder.SetFields(fields)
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        expected, _ = expectedExpr.__teal__(avm5Options)
        expected.addIncoming()
        expected = pt.TealBlock.NormalizeBlocks(expected)

        actual, _ = expr.__teal__(avm5Options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreScratchSlotEquality(), pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected

        assert pt.TealBlock.MatchScratchSlotReferences(
            pt.TealBlock.GetReferencedScratchSlots(actual),
            pt.TealBlock.GetReferencedScratchSlots(expected),
        )

        if len(fields) != 0:
            with pytest.raises(pt.TealInputError):
                expr.__teal__(avm4Options)


def test_InnerTxnBuilder_Execute():
    for fields, expectedExpr in ITXN_FIELDS_CASES:
        expr = pt.InnerTxnBuilder.Execute(fields)

        expected, _ = pt.Seq(
            pt.InnerTxnBuilder.Begin(),
            expectedExpr,
            pt.InnerTxnBuilder.Submit(),
        ).__teal__(avm5Options)
        expected.addIncoming()
        expected = pt.TealBlock.NormalizeBlocks(expected)

        actual, _ = expr.__teal__(avm5Options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreScratchSlotEquality(), pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected

        assert pt.TealBlock.MatchScratchSlotReferences(
            pt.TealBlock.GetReferencedScratchSlots(actual),
            pt.TealBlock.GetReferencedScratchSlots(expected),
        )

        with pytest.raises(pt.TealInputError):
            expr.__teal__(avm4Options)


ITXN_METHOD_CASES = (
    (
        pt.Int(1),
        "add(uint64,uint64)void",
        [t1_1 := pt.Itob(pt.Int(1)), t1_2 := pt.Itob(pt.Int(1))],
        {TxnField.fee: pt.Int(0)},
        pt.Seq(
            pt.InnerTxnBuilder.SetFields(
                {
                    pt.TxnField.type_enum: TxnType.ApplicationCall,
                    pt.TxnField.application_id: pt.Int(1),
                    pt.TxnField.application_args: [
                        pt.MethodSignature("add(uint64,uint64)void"),
                        t1_1,
                        t1_2,
                    ],
                    pt.TxnField.fee: pt.Int(0),
                }
            ),
        ),
        None,
    ),
    (
        pt.Int(1),
        "add(uint64,uint64)void",
        [t2_1 := pt.abi.Uint64(), t2_2 := pt.abi.Uint64()],
        {TxnField.fee: pt.Int(0)},
        pt.Seq(
            pt.InnerTxnBuilder.SetFields(
                {
                    pt.TxnField.type_enum: TxnType.ApplicationCall,
                    pt.TxnField.application_id: pt.Int(1),
                    pt.TxnField.application_args: [
                        pt.MethodSignature("add(uint64,uint64)void"),
                        t2_1.encode(),
                        t2_2.encode(),
                    ],
                    pt.TxnField.fee: pt.Int(0),
                }
            ),
        ),
        None,
    ),
    (
        pt.Int(1),
        "add(application,account,asset)void",
        [
            t3_1 := pt.abi.Application(),
            t3_2 := pt.abi.Account(),
            t3_3 := pt.abi.Asset(),
        ],
        {TxnField.fee: pt.Int(0)},
        pt.Seq(
            pt.InnerTxnBuilder.SetFields(
                {
                    pt.TxnField.type_enum: TxnType.ApplicationCall,
                    pt.TxnField.application_id: pt.Int(1),
                    pt.TxnField.accounts: [t3_2.address()],
                    pt.TxnField.applications: [t3_1.application_id()],
                    pt.TxnField.assets: [t3_3.asset_id()],
                    pt.TxnField.application_args: [
                        pt.MethodSignature("add(application,account,asset)void"),
                        pt.Bytes(b"\x01"),
                        pt.Bytes(b"\x01"),
                        pt.Bytes(b"\x00"),
                    ],
                    pt.TxnField.fee: pt.Int(0),
                }
            ),
        ),
        None,
    ),
    (
        pt.Int(1),
        "add(application,account,asset)void",
        [
            t4_1 := pt.Int(1),
            t4_2 := pt.Global.zero_address(),
            t4_3 := pt.Int(2),
        ],
        {TxnField.fee: pt.Int(0)},
        pt.Seq(
            pt.InnerTxnBuilder.SetFields(
                {
                    pt.TxnField.type_enum: TxnType.ApplicationCall,
                    pt.TxnField.application_id: pt.Int(1),
                    pt.TxnField.accounts: [t4_2],
                    pt.TxnField.applications: [t4_1],
                    pt.TxnField.assets: [t4_3],
                    pt.TxnField.application_args: [
                        pt.MethodSignature("add(application,account,asset)void"),
                        pt.Bytes(b"\x01"),
                        pt.Bytes(b"\x01"),
                        pt.Bytes(b"\x00"),
                    ],
                    pt.TxnField.fee: pt.Int(0),
                }
            ),
        ),
        None,
    ),
    (
        pt.Int(1),
        "add(pay,txn,appl)void",
        [
            t5_1 := {TxnField.type_enum: TxnType.Payment},
            t5_2 := {TxnField.type_enum: TxnType.AssetTransfer},
            t5_3 := {TxnField.type_enum: TxnType.ApplicationCall},
        ],
        {TxnField.fee: pt.Int(0)},
        pt.Seq(
            pt.InnerTxnBuilder.SetFields(t5_1),  # type: ignore
            pt.InnerTxnBuilder.Next(),
            pt.InnerTxnBuilder.SetFields(t5_2),  # type: ignore
            pt.InnerTxnBuilder.Next(),
            pt.InnerTxnBuilder.SetFields(t5_3),  # type: ignore
            pt.InnerTxnBuilder.Next(),
            pt.InnerTxnBuilder.SetFields(
                {
                    pt.TxnField.type_enum: TxnType.ApplicationCall,
                    pt.TxnField.application_id: pt.Int(1),
                    pt.TxnField.application_args: [
                        pt.MethodSignature("add(pay,txn,appl)void"),
                    ],
                    pt.TxnField.fee: pt.Int(0),
                }
            ),
        ),
        None,
    ),
    (
        pt.Int(1),
        "query(byte[],uint64)void",
        [t6_1 := pt.abi.DynamicBytes(), t6_2 := pt.abi.Uint64()],
        {TxnField.fee: pt.Int(0)},
        pt.Seq(
            pt.InnerTxnBuilder.SetFields(
                {
                    pt.TxnField.type_enum: TxnType.ApplicationCall,
                    pt.TxnField.application_id: pt.Int(1),
                    pt.TxnField.application_args: [
                        pt.MethodSignature("query(byte[],uint64)void"),
                        t6_1.encode(),
                        t6_2.encode(),
                    ],
                    pt.TxnField.fee: pt.Int(0),
                }
            ),
        ),
        None,
    ),
    # Error cases
    (
        pt.Int(1),
        "add(pay,txn,appl)void",
        [
            {},
            {TxnField.type_enum: TxnType.AssetTransfer},
            {TxnField.type_enum: TxnType.ApplicationCall},
        ],
        None,
        None,
        pt.TealInputError,
    ),
    (
        pt.Int(1),
        "add(pay,txn,appl)void",
        [
            {TxnField.type_enum: pt.Int(10)},
            {TxnField.type_enum: TxnType.AssetTransfer},
            {TxnField.type_enum: TxnType.ApplicationCall},
        ],
        None,
        None,
        pt.TealTypeError,
    ),
    (
        pt.Int(1),
        "add(pay,txn,appl)void",
        [
            {TxnField.type_enum: TxnType.ApplicationCall},
            {TxnField.type_enum: TxnType.AssetTransfer},
            {TxnField.type_enum: TxnType.ApplicationCall},
        ],
        None,
        None,
        pt.TealInputError,
    ),
    (
        pt.Int(1),
        "add(application,account,asset)void",
        [
            pt.abi.Asset(),
            pt.abi.Account(),
            pt.abi.Asset(),
        ],
        None,
        None,
        pt.TealTypeError,
    ),
    (
        pt.Int(1),
        "add(application)void",
        [
            pt.Bytes(""),
        ],
        None,
        None,
        pt.TealTypeError,
    ),
    (
        pt.Int(1),
        "add(asset)void",
        [
            pt.Bytes(""),
        ],
        None,
        None,
        pt.TealTypeError,
    ),
    (
        pt.Int(1),
        "add(account)void",
        [
            pt.Int(1),
        ],
        None,
        None,
        pt.TealTypeError,
    ),
    (
        pt.Int(1),
        "add(uint64,uint64)void",
        [pt.abi.String(), pt.abi.Uint64()],
        None,
        None,
        pt.TealTypeError,
    ),
    (
        pt.Int(1),
        "add(uint64,uint64)void",
        [pt.abi.Uint64()],
        None,
        None,
        pt.TealInputError,
    ),
    (
        pt.Int(1),
        "add(uint64,uint64)void",
        [pt.abi.Uint64(), pt.abi.Uint64(), pt.abi.Uint64()],
        None,
        None,
        pt.TealInputError,
    ),
)


@pytest.mark.parametrize(
    "app_id, sig, args, extra_fields, expected_expr, expected_error", ITXN_METHOD_CASES
)
def test_InnerTxnBuilder_method_call(
    app_id: pt.Expr,
    sig: str,
    args: list[pt.abi.BaseType | pt.Expr | dict[pt.TxnField, pt.Expr | list[pt.Expr]]],
    extra_fields: dict[pt.TxnField, pt.Expr | list[pt.Expr]],
    expected_expr: pt.Expr,
    expected_error: type[Exception],
):

    if expected_error is not None:
        with pytest.raises(expected_error):
            pt.InnerTxnBuilder.MethodCall(
                app_id=app_id,
                method_signature=sig,
                args=args,
                extra_fields=extra_fields,
            )
        with pytest.raises(expected_error):
            pt.InnerTxnBuilder.ExecuteMethodCall(
                app_id=app_id,
                method_signature=sig,
                args=args,
                extra_fields=extra_fields,
            )
        return

    # First run the test with MethodCall
    expr: pt.Expr = pt.InnerTxnBuilder.MethodCall(
        app_id=app_id, method_signature=sig, args=args, extra_fields=extra_fields
    )
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected, _ = expected_expr.__teal__(avm6Options)
    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.__teal__(avm6Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreScratchSlotEquality(), pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected
        assert pt.TealBlock.MatchScratchSlotReferences(
            pt.TealBlock.GetReferencedScratchSlots(actual),
            pt.TealBlock.GetReferencedScratchSlots(expected),
        )

    # Now run the same test with ExecuteMethodCall
    expr = pt.InnerTxnBuilder.ExecuteMethodCall(
        app_id=app_id, method_signature=sig, args=args, extra_fields=extra_fields
    )
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected, _ = pt.Seq(
        pt.InnerTxnBuilder.Begin(), expected_expr, pt.InnerTxnBuilder.Submit()
    ).__teal__(avm6Options)
    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.__teal__(avm6Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreScratchSlotEquality(), pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected
        assert pt.TealBlock.MatchScratchSlotReferences(
            pt.TealBlock.GetReferencedScratchSlots(actual),
            pt.TealBlock.GetReferencedScratchSlots(expected),
        )


# txn_test.py performs additional testing



from typing import TYPE_CHECKING, Any

from pyteal.ast.expr import Expr
from pyteal.pragma import is_valid_compiler_version, pragma

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Pragma(Expr):
    """A meta expression which defines a pragma for a specific subsection of PyTeal code.

    This expression does not affect the underlying compiled TEAL code in any way."""

    def __init__(self, child: Expr, *, compiler_version: str, **kwargs: Any) -> None:
        """Define a pragma for a specific subsection of PyTeal code.

        The Pragma expression does not affect the underlying compiled TEAL code in any way,
        it merely sets a pragma for the underlying expression.

        Args:
            child: The expression to wrap.
            compiler_version: Acceptable versions of the compiler. Will fail if the current PyTeal version
                is not contained in the range. Follows the npm `semver range scheme <https://github.com/npm/node-semver#ranges>`_
                for specifying compatible versions.

        For example:

        .. code-block:: python

            @Subroutine(TealType.uint64)
            def example() -> Expr:
                # this will fail during compilation if the current PyTeal version does not satisfy
                # the version constraint
                return Pragma(
                    Seq(...),
                    compiler_version="^0.14.0"
                )
        """
        super().__init__()

        self.child = child

        if not is_valid_compiler_version(compiler_version):
            raise ValueError("Invalid compiler version: {}".format(compiler_version))
        self.compiler_version = compiler_version

    def __teal__(self, options: "CompileOptions"):
        pragma(compiler_version=self.compiler_version)

        return self.child.__teal__(options)

    def __str__(self):
        return "(pragma {})".format(self.child)

    def type_of(self):
        return self.child.type_of()

    def has_return(self):
        return self.child.has_return()


Pragma.__module__ = "pyteal"



import pytest

import pyteal as pt

avm6Options = pt.CompileOptions(version=6)
avm7Options = pt.CompileOptions(version=7)


def test_block_seed():
    arg = pt.Int(0)
    expr = pt.Block.seed(arg)
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.block, "BlkSeed"),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


def test_block_seed_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Block.seed(pt.Bytes(""))


def test_block_timestamp():
    arg = pt.Int(0)
    expr = pt.Block.timestamp(arg)
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.block, "BlkTimestamp"),
        ]
    )

    actual, _ = expr.__teal__(avm7Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm6Options)


def test_block_timestamp_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Block.timestamp(pt.Txn.sender())



from pyteal.ast.abi.string import String, StringTypeSpec
from pyteal.ast.abi.address import (
    AddressTypeSpec,
    Address,
    AddressLength,
)
from pyteal.ast.abi.type import TypeSpec, BaseType, ComputedValue, ReturnedValue
from pyteal.ast.abi.bool import BoolTypeSpec, Bool
from pyteal.ast.abi.uint import (
    UintTypeSpec,
    Uint,
    ByteTypeSpec,
    Byte,
    Uint8TypeSpec,
    Uint8,
    Uint16TypeSpec,
    Uint16,
    Uint32TypeSpec,
    Uint32,
    Uint64TypeSpec,
    Uint64,
)
from pyteal.ast.abi.tuple import (
    TupleTypeSpec,
    Tuple,
    TupleElement,
    Tuple0,
    Tuple1,
    Tuple2,
    Tuple3,
    Tuple4,
    Tuple5,
    NamedTuple,
    NamedTupleTypeSpec,
    Field,
)
from pyteal.ast.abi.array_base import ArrayTypeSpec, Array, ArrayElement
from pyteal.ast.abi.array_static import (
    StaticArrayTypeSpec,
    StaticArray,
    StaticBytesTypeSpec,
    StaticBytes,
)
from pyteal.ast.abi.array_dynamic import (
    DynamicArrayTypeSpec,
    DynamicArray,
    DynamicBytesTypeSpec,
    DynamicBytes,
)
from pyteal.ast.abi.reference_type import (
    ReferenceTypeSpec,
    ReferenceType,
    Account,
    AccountTypeSpec,
    Asset,
    AssetTypeSpec,
    Application,
    ApplicationTypeSpec,
    ReferenceTypeSpecs,
)
from pyteal.ast.abi.transaction import (
    Transaction,
    TransactionTypeSpec,
    PaymentTransaction,
    PaymentTransactionTypeSpec,
    ApplicationCallTransaction,
    ApplicationCallTransactionTypeSpec,
    AssetConfigTransaction,
    AssetConfigTransactionTypeSpec,
    AssetFreezeTransaction,
    AssetFreezeTransactionTypeSpec,
    AssetTransferTransaction,
    AssetTransferTransactionTypeSpec,
    KeyRegisterTransaction,
    KeyRegisterTransactionTypeSpec,
    TransactionTypeSpecs,
)
from pyteal.ast.abi.method_return import MethodReturn
from pyteal.ast.abi.util import (
    algosdk_from_annotation,
    algosdk_from_type_spec,
    make,
    size_of,
    type_spec_from_annotation,
    type_spec_from_algosdk,
    type_specs_from_signature,
    contains_type_spec,
)

__all__ = [
    "String",
    "StringTypeSpec",
    "Account",
    "AccountTypeSpec",
    "Asset",
    "ReferenceTypeSpec",
    "ReferenceType",
    "AssetTypeSpec",
    "Application",
    "ApplicationTypeSpec",
    "ReferenceType",
    "ReferenceTypeSpec",
    "ReferenceTypeSpecs",
    "Address",
    "AddressTypeSpec",
    "AddressLength",
    "TypeSpec",
    "BaseType",
    "ComputedValue",
    "ReturnedValue",
    "BoolTypeSpec",
    "Bool",
    "UintTypeSpec",
    "Uint",
    "ByteTypeSpec",
    "Byte",
    "Uint8TypeSpec",
    "Uint8",
    "Uint16TypeSpec",
    "Uint16",
    "Uint32TypeSpec",
    "Uint32",
    "Uint64TypeSpec",
    "Uint64",
    "TupleTypeSpec",
    "Tuple",
    "TupleElement",
    "Tuple0",
    "Tuple1",
    "Tuple2",
    "Tuple3",
    "Tuple4",
    "Tuple5",
    "NamedTuple",
    "NamedTupleTypeSpec",
    "Field",
    "ArrayTypeSpec",
    "Array",
    "ArrayElement",
    "StaticArrayTypeSpec",
    "StaticArray",
    "StaticBytesTypeSpec",
    "StaticBytes",
    "DynamicArrayTypeSpec",
    "DynamicArray",
    "DynamicBytesTypeSpec",
    "DynamicBytes",
    "MethodReturn",
    "Transaction",
    "TransactionTypeSpec",
    "PaymentTransaction",
    "PaymentTransactionTypeSpec",
    "ApplicationCallTransaction",
    "ApplicationCallTransactionTypeSpec",
    "AssetConfigTransaction",
    "AssetConfigTransactionTypeSpec",
    "AssetFreezeTransaction",
    "AssetFreezeTransactionTypeSpec",
    "AssetTransferTransaction",
    "AssetTransferTransactionTypeSpec",
    "KeyRegisterTransaction",
    "KeyRegisterTransactionTypeSpec",
    "TransactionTypeSpecs",
    "type_spec_from_annotation",
    "type_spec_from_algosdk",
    "type_specs_from_signature",
    "make",
    "size_of",
    "algosdk_from_annotation",
    "algosdk_from_type_spec",
    "contains_type_spec",
]



import pytest

import pyteal as pt


def test_addr():
    expr = pt.Addr("NJUWK3DJNZTWU2LFNRUW4Z3KNFSWY2LOM5VGSZLMNFXGO2TJMVWGS3THMF")
    assert expr.type_of() == pt.TealType.bytes
    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(
                expr,
                pt.Op.addr,
                "NJUWK3DJNZTWU2LFNRUW4Z3KNFSWY2LOM5VGSZLMNFXGO2TJMVWGS3THMF",
            )
        ]
    )
    actual, _ = expr.__teal__(pt.CompileOptions())
    assert actual == expected


def test_addr_invalid():
    with pytest.raises(pt.TealInputError):
        pt.Addr("NJUWK3DJNZTWU2LFNRUW4Z3KNFSWY2LOM5VGSZLMNFXGO2TJMVWGS3TH")

    with pytest.raises(pt.TealInputError):
        pt.Addr("000000000000000000000000000000000000000000000000000000000")

    with pytest.raises(pt.TealInputError):
        pt.Addr(2)



from typing import Union, List, Optional, TYPE_CHECKING

from pyteal.ir.tealcomponent import TealComponent
from pyteal.ir.labelref import LabelReference
from pyteal.ir.ops import Op
from pyteal.errors import TealInternalError

if TYPE_CHECKING:
    from pyteal.ast import Expr, ScratchSlot, SubroutineDefinition


class TealOp(TealComponent):
    def __init__(
        self,
        expr: Optional["Expr"],
        op: Op,
        *args: Union[int, str, LabelReference, "ScratchSlot", "SubroutineDefinition"]
    ) -> None:
        super().__init__(expr)
        self.op = op
        self.args = list(args)

    def getOp(self) -> Op:
        return self.op

    def getSlots(self) -> List["ScratchSlot"]:
        from pyteal.ast import ScratchSlot

        return [arg for arg in self.args if isinstance(arg, ScratchSlot)]

    def assignSlot(self, slot: "ScratchSlot", location: int) -> None:
        for i, arg in enumerate(self.args):
            if slot == arg:
                self.args[i] = location

    def getSubroutines(self) -> List["SubroutineDefinition"]:
        from pyteal.ast import SubroutineDefinition

        return [arg for arg in self.args if isinstance(arg, SubroutineDefinition)]

    def resolveSubroutine(self, subroutine: "SubroutineDefinition", label: str) -> None:
        for i, arg in enumerate(self.args):
            if subroutine == arg:
                self.args[i] = label

    def assemble(self) -> str:
        from pyteal.ast import ScratchSlot, SubroutineDefinition

        parts = [str(self.op)]
        for arg in self.args:
            if isinstance(arg, ScratchSlot):
                raise TealInternalError("Slot not assigned: {}".format(arg))

            if isinstance(arg, SubroutineDefinition):
                raise TealInternalError("Subroutine not resolved: {}".format(arg))

            if isinstance(arg, int):
                parts.append(str(arg))
            elif isinstance(arg, LabelReference):
                parts.append(arg.getLabel())
            else:
                parts.append(arg)

        return " ".join(parts)

    def __repr__(self) -> str:
        args = [str(self.op)]
        for a in self.args:
            args.append(repr(a))

        return "TealOp({})".format(", ".join(args))

    def __hash__(self) -> int:
        return (self.op, *self.args).__hash__()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TealOp):
            return False

        if TealComponent.Context.checkExprEquality and self.expr is not other.expr:
            return False

        if not TealComponent.Context.checkScratchSlotEquality:
            from pyteal import ScratchSlot

            if len(self.args) != len(other.args):
                return False
            for myArg, otherArg in zip(self.args, other.args):
                if type(myArg) is ScratchSlot and type(otherArg) is ScratchSlot:
                    continue
                if myArg != otherArg:
                    return False
        elif self.args != other.args:
            return False

        return self.op == other.op


TealOp.__module__ = "pyteal"



import pytest
from typing import List

import pyteal as pt

options = pt.CompileOptions()


def __test_single(expr: pt.MultiValue):
    assert expr.output_slots[0] != expr.output_slots[1]

    with pt.TealComponent.Context.ignoreExprEquality():
        assert expr.output_slots[0].load().__teal__(options) == pt.ScratchLoad(
            expr.output_slots[0]
        ).__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert expr.output_slots[1].load().__teal__(options) == pt.ScratchLoad(
            expr.output_slots[1]
        ).__teal__(options)

    assert expr.type_of() == pt.TealType.none


def __test_single_conditional(
    expr: pt.MultiValue, op, args: List[pt.Expr], iargs, reducer
):
    __test_single(expr)

    expected_call = pt.TealSimpleBlock(
        [
            pt.TealOp(expr, op, *iargs),
            pt.TealOp(expr.output_slots[1].store(), pt.Op.store, expr.output_slots[1]),
            pt.TealOp(expr.output_slots[0].store(), pt.Op.store, expr.output_slots[0]),
        ]
    )

    ifExpr = (
        pt.If(expr.output_slots[1].load())
        .Then(expr.output_slots[0].load())
        .Else(pt.App.globalGet(pt.Bytes("None")))
    )
    ifBlockStart, _ = ifExpr.__teal__(options)

    expected_call.setNextBlock(ifBlockStart)

    if len(args) == 0:
        expected: pt.TealBlock = expected_call
    elif len(args) == 1:
        expected, after_arg = args[0].__teal__(options)
        after_arg.setNextBlock(expected_call)
    elif len(args) == 2:
        expected, after_arg_1 = args[0].__teal__(options)
        arg_2, after_arg_2 = args[1].__teal__(options)
        after_arg_1.setNextBlock(arg_2)
        after_arg_2.setNextBlock(expected_call)

    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.outputReducer(reducer).__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def __test_single_assert(expr: pt.MultiValue, op, args: List[pt.Expr], iargs, reducer):
    __test_single(expr)

    expected_call = pt.TealSimpleBlock(
        [
            pt.TealOp(expr, op, *iargs),
            pt.TealOp(expr.output_slots[1].store(), pt.Op.store, expr.output_slots[1]),
            pt.TealOp(expr.output_slots[0].store(), pt.Op.store, expr.output_slots[0]),
        ]
    )

    assertExpr = pt.Seq(
        pt.Assert(expr.output_slots[1].load()), expr.output_slots[0].load()
    )
    assertBlockStart, _ = assertExpr.__teal__(options)

    expected_call.setNextBlock(assertBlockStart)

    if len(args) == 0:
        expected: pt.TealBlock = expected_call
    elif len(args) == 1:
        expected, after_arg = args[0].__teal__(options)
        after_arg.setNextBlock(expected_call)
    elif len(args) == 2:
        expected, after_arg_1 = args[0].__teal__(options)
        arg_2, after_arg_2 = args[1].__teal__(options)
        after_arg_1.setNextBlock(arg_2)
        after_arg_2.setNextBlock(expected_call)

    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.outputReducer(reducer).__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def __test_single_with_vars(
    expr: pt.MultiValue, op, args: List[pt.Expr], iargs, var1, var2, reducer
):
    __test_single(expr)

    expected_call = pt.TealSimpleBlock(
        [
            pt.TealOp(expr, op, *iargs),
            pt.TealOp(expr.output_slots[1].store(), pt.Op.store, expr.output_slots[1]),
            pt.TealOp(expr.output_slots[0].store(), pt.Op.store, expr.output_slots[0]),
        ]
    )

    varExpr = pt.Seq(
        var1.store(expr.output_slots[1].load()), var2.store(expr.output_slots[0].load())
    )
    varBlockStart, _ = varExpr.__teal__(options)

    expected_call.setNextBlock(varBlockStart)

    if len(args) == 0:
        expected: pt.TealBlock = expected_call
    elif len(args) == 1:
        expected, after_arg = args[0].__teal__(options)
        after_arg.setNextBlock(expected_call)
    elif len(args) == 2:
        expected, after_arg_1 = args[0].__teal__(options)
        arg_2, after_arg_2 = args[1].__teal__(options)
        after_arg_1.setNextBlock(arg_2)
        after_arg_2.setNextBlock(expected_call)

    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.outputReducer(reducer).__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


@pytest.mark.parametrize(
    "op",
    [
        pt.Op.app_global_get_ex,
        pt.Op.app_local_get_ex,
        pt.Op.asset_holding_get,
        pt.Op.asset_params_get,
    ],
)
@pytest.mark.parametrize(
    "type", [pt.TealType.uint64, pt.TealType.bytes, pt.TealType.anytype]
)
@pytest.mark.parametrize("iargs", [[], ["AssetFrozen"]])
@pytest.mark.parametrize("args", [[], [pt.Int(0)], [pt.Int(1), pt.Int(2)]])
def test_multi_value(op, type, iargs, args):
    reducer = (
        lambda value, hasValue: pt.If(hasValue)
        .Then(value)
        .Else(pt.App.globalGet(pt.Bytes("None")))
    )
    expr = pt.MultiValue(
        op, [type, pt.TealType.uint64], immediate_args=iargs, args=args
    )
    __test_single_conditional(expr, op, args, iargs, reducer)

    reducer = lambda value, hasValue: pt.Seq(pt.Assert(hasValue), value)  # noqa: E731
    expr = pt.MultiValue(
        op, [type, pt.TealType.uint64], immediate_args=iargs, args=args
    )
    __test_single_assert(expr, op, args, iargs, reducer)

    hasValueVar = pt.ScratchVar(pt.TealType.uint64)
    valueVar = pt.ScratchVar(type)
    reducer = lambda value, hasValue: pt.Seq(  # noqa: E731
        hasValueVar.store(hasValue), valueVar.store(value)
    )
    expr = pt.MultiValue(
        op, [type, pt.TealType.uint64], immediate_args=iargs, args=args
    )
    __test_single_with_vars(expr, op, args, iargs, hasValueVar, valueVar, reducer)


def test_multi_compile_check():
    def never_fails(options):
        return

    program_never_fails = pt.MultiValue(
        pt.Op.app_global_get_ex,
        [pt.TealType.uint64, pt.TealType.uint64],
        compile_check=never_fails,
    )
    program_never_fails.__teal__(options)

    class TestException(Exception):
        pass

    def always_fails(options):
        raise TestException()

    program_always_fails = pt.MultiValue(
        pt.Op.app_global_get_ex,
        [pt.TealType.uint64, pt.TealType.uint64],
        compile_check=always_fails,
    )

    with pytest.raises(TestException):
        program_always_fails.__teal__(options)



from typing import TYPE_CHECKING, Optional
from pyteal.ast.seq import _use_seq_if_multiple

from pyteal.types import TealType, require_type
from pyteal.ir import TealSimpleBlock, TealConditionalBlock
from pyteal.errors import TealCompileError
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class For(Expr):
    """For expression."""

    def __init__(self, start: Expr, cond: Expr, step: Expr) -> None:
        """Create a new For expression.

        When this For expression is executed, the condition will be evaluated, and if it produces a
        true value, doBlock will be executed and return to the start of the expression execution.
        Otherwise, no branch will be executed.

        Args:
            start: Expression setting the variable's initial value
            cond: The condition to check. Must evaluate to uint64.
            step: Expression to update the variable's value.

        Example:
            .. code-block:: python

                i = ScratchVar()
                For(i.store(Int(0)), i.load() < Int(10), i.store(i.load() + Int(1))
                    .Do(expr1, expr2, ...)
        """
        super().__init__()
        require_type(cond, TealType.uint64)
        require_type(start, TealType.none)
        require_type(step, TealType.none)

        self.start = start
        self.cond = cond
        self.step = step
        self.doBlock: Optional[Expr] = None

    def __teal__(self, options: "CompileOptions"):
        if self.doBlock is None:
            raise TealCompileError("For expression must have a doBlock", self)

        options.enterLoop()

        end = TealSimpleBlock([])
        start, startEnd = self.start.__teal__(options)
        condStart, condEnd = self.cond.__teal__(options)
        doStart, doEnd = self.doBlock.__teal__(options)

        stepStart, stepEnd = self.step.__teal__(options)
        stepEnd.setNextBlock(condStart)
        doEnd.setNextBlock(stepStart)

        branchBlock = TealConditionalBlock([])
        branchBlock.setTrueBlock(doStart)
        branchBlock.setFalseBlock(end)

        condEnd.setNextBlock(branchBlock)

        startEnd.setNextBlock(condStart)

        breakBlocks, continueBlocks = options.exitLoop()

        for block in breakBlocks:
            block.setNextBlock(end)

        for block in continueBlocks:
            block.setNextBlock(stepStart)

        return start, end

    def __str__(self):
        if self.start is None:
            raise TealCompileError("For expression must have a start", self)
        if self.cond is None:
            raise TealCompileError("For expression must have a condition", self)
        if self.step is None:
            raise TealCompileError("For expression must have a end", self)
        if self.doBlock is None:
            raise TealCompileError("For expression must have a doBlock", self)

        return "(For {} {} {} {})".format(
            self.start, self.cond, self.step, self.doBlock
        )

    def type_of(self):
        if self.doBlock is None:
            raise TealCompileError("For expression must have a doBlock", self)
        return TealType.none

    def has_return(self):
        return False

    def Do(self, doBlock: Expr, *do_block_multi: Expr):
        if self.doBlock is not None:
            raise TealCompileError("For expression already has a doBlock", self)

        doBlock = _use_seq_if_multiple(doBlock, *do_block_multi)

        require_type(doBlock, TealType.none)
        self.doBlock = doBlock
        return self


For.__module__ = "pyteal"



from typing import TypeVar, Union, Sequence, Callable

from pyteal.types import TealType
from pyteal.errors import TealInputError
from pyteal.ast.expr import Expr
from pyteal.ast.int import Int
from pyteal.ast.bytes import Bytes
from pyteal.ast.unaryexpr import Not
from pyteal.ast.binaryexpr import GetBit
from pyteal.ast.ternaryexpr import SetBit
from pyteal.ast.abi.type import ComputedValue, TypeSpec, BaseType
from pyteal.ast.abi.uint import NUM_BITS_IN_BYTE


class BoolTypeSpec(TypeSpec):
    def new_instance(self) -> "Bool":
        return Bool()

    def annotation_type(self) -> "type[Bool]":
        return Bool

    def is_dynamic(self) -> bool:
        # Only accurate if this value is alone, since up to 8 consecutive bools will fit into a single byte
        return False

    def byte_length_static(self) -> int:
        return 1

    def storage_type(self) -> TealType:
        return TealType.uint64

    def __eq__(self, other: object) -> bool:
        return isinstance(other, BoolTypeSpec)

    def __str__(self) -> str:
        return "bool"


BoolTypeSpec.__module__ = "pyteal.abi"


class Bool(BaseType):
    def __init__(self) -> None:
        super().__init__(BoolTypeSpec())

    def get(self) -> Expr:
        """Return the value held by this Bool as a PyTeal expression.

        If the held value is true, an expression that evaluates to 1 will be returned. Otherwise, an
        expression that evaluates to 0 will be returned. In either case, the expression will have the
        type TealType.uint64.
        """
        return self.stored_value.load()

    def set(self, value: Union[bool, Expr, "Bool", ComputedValue["Bool"]]) -> Expr:
        """Set the value of this Bool to the input value.

        The behavior of this method depends on the input argument type:

            * :code:`bool`: set the value to a Python boolean value.
            * :code:`Expr`: set the value to the result of a PyTeal expression, which must evaluate to a TealType.uint64. All values greater than 0 are considered true, while 0 is considered false.
            * :code:`Bool`: copy the value from another Bool.
            * :code:`ComputedValue[Bool]`: copy the value from a Bool produced by a ComputedValue.

        Args:
            value: The new value this Bool should take. This must follow the above constraints.

        Returns:
            An expression which stores the given value into this Bool.
        """
        if isinstance(value, ComputedValue):
            return self._set_with_computed_type(value)

        checked = False
        if type(value) is bool:
            value = Int(1 if value else 0)
            checked = True

        if isinstance(value, BaseType):
            if value.type_spec() != self.type_spec():
                raise TealInputError(
                    "Cannot set type bool to {}".format(value.type_spec())
                )
            value = value.get()
            checked = True

        if checked:
            return self.stored_value.store(value)

        # Not(Not(value)) coerces all values greater than 0 to 1
        return self.stored_value.store(Not(Not(value)))

    def decode(
        self,
        encoded: Expr,
        *,
        start_index: Expr = None,
        end_index: Expr = None,
        length: Expr = None
    ) -> Expr:
        if start_index is None:
            start_index = Int(0)
        return self.decode_bit(encoded, start_index * Int(NUM_BITS_IN_BYTE))

    def decode_bit(self, encoded, bit_index: Expr) -> Expr:
        return self.stored_value.store(GetBit(encoded, bit_index))

    def encode(self) -> Expr:
        return SetBit(Bytes(b"\x00"), Int(0), self.get())


Bool.__module__ = "pyteal.abi"


def _bool_aware_static_byte_length(types: Sequence[TypeSpec]) -> int:
    length = 0
    ignoreNext = 0
    for i, t in enumerate(types):
        if ignoreNext > 0:
            ignoreNext -= 1
            continue
        if t == BoolTypeSpec():
            numBools = _consecutive_bool_type_spec_num(types, i)
            ignoreNext = numBools - 1
            length += _bool_sequence_length(numBools)
            continue
        length += t.byte_length_static()
    return length


T = TypeVar("T")


def _consecutive_thing_num(
    things: Sequence[T], start_index: int, condition: Callable[[T], bool]
) -> int:
    numConsecutiveThings = 0
    for t in things[start_index:]:
        if not condition(t):
            break
        numConsecutiveThings += 1
    return numConsecutiveThings


def _consecutive_bool_type_spec_num(types: Sequence[TypeSpec], start_index: int) -> int:
    if len(types) != 0 and not isinstance(types[0], TypeSpec):
        raise TypeError("Sequence of types expected")
    return _consecutive_thing_num(types, start_index, lambda t: t == BoolTypeSpec())


def _consecutive_bool_instance_num(values: Sequence[BaseType], start_index: int) -> int:
    if len(values) != 0 and not isinstance(values[0], BaseType):
        raise TypeError(
            "Sequence of types expected, but got {}".format(type(values[0]))
        )
    return _consecutive_thing_num(
        values, start_index, lambda t: t.type_spec() == BoolTypeSpec()
    )


def _bool_sequence_length(num_bools: int) -> int:
    """Get the length in bytes of an encoding of `num_bools` consecutive booleans values."""
    return (num_bools + NUM_BITS_IN_BYTE - 1) // NUM_BITS_IN_BYTE


def _encode_bool_sequence(values: Sequence[Bool]) -> Expr:
    """Encoding a sequences of boolean values into a byte string.

    Args:
        values: The values to encode. Each must be an instance of Bool.

    Returns:
        An expression which creates an encoded byte string with the input boolean values.
    """
    length = _bool_sequence_length(len(values))
    expr: Expr = Bytes(b"\x00" * length)

    for i, value in enumerate(values):
        expr = SetBit(expr, Int(i), value.get())

    return expr



import pyteal as pt
from pyteal.ast.maybe_test import assert_MaybeValue_equality

options = pt.CompileOptions()
avm4Options = pt.CompileOptions(version=4)
avm5Options = pt.CompileOptions(version=5)
avm6Options = pt.CompileOptions(version=6)


def test_acct_param_balance_valid():
    arg = pt.Int(1)
    expr = pt.AccountParam.balance(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.acct_params_get, "AcctBalance"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm6Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_acct_param_min_balance_valid():
    arg = pt.Int(0)
    expr = pt.AccountParam.minBalance(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.acct_params_get, "AcctMinBalance"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm6Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_acct_param_auth_addr_valid():
    arg = pt.Int(1)
    expr = pt.AccountParam.authAddr(arg)
    assert expr.type_of() == pt.TealType.none
    assert expr.value().type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(arg, pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.acct_params_get, "AcctAuthAddr"),
            pt.TealOp(None, pt.Op.store, expr.slotOk),
            pt.TealOp(None, pt.Op.store, expr.slotValue),
        ]
    )

    actual, _ = expr.__teal__(avm6Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_AccountParamObject():
    for account in (
        pt.Int(7),
        pt.Addr("QSA6K5MNJPEGO5SDSWXBM3K4UEI3Q2NCPS2OUXVJI5QPCHMVI27MFRSHKI"),
    ):
        obj = pt.AccountParamObject(account)

        assert obj._account is account

        assert_MaybeValue_equality(
            obj.balance(), pt.AccountParam.balance(account), avm6Options
        )
        assert_MaybeValue_equality(
            obj.min_balance(), pt.AccountParam.minBalance(account), avm6Options
        )
        assert_MaybeValue_equality(
            obj.auth_address(), pt.AccountParam.authAddr(account), avm6Options
        )



import pyteal as pt

from pyteal.compiler.constants import (
    extractIntValue,
    extractBytesValue,
    extractAddrValue,
    createConstantBlocks,
    extractMethodSigValue,
)


def test_extractIntValue():
    tests = [
        (pt.TealOp(None, pt.Op.int, 0), 0),
        (pt.TealOp(None, pt.Op.int, 5), 5),
        (pt.TealOp(None, pt.Op.int, "pay"), 1),
        (pt.TealOp(None, pt.Op.int, "NoOp"), 0),
        (pt.TealOp(None, pt.Op.int, "UpdateApplication"), 4),
        (pt.TealOp(None, pt.Op.int, "TMPL_NAME"), "TMPL_NAME"),
    ]

    for op, expected in tests:
        actual = extractIntValue(op)
        assert actual == expected


def test_extractBytesValue():
    tests = [
        (pt.TealOp(None, pt.Op.byte, '""'), b""),
        (pt.TealOp(None, pt.Op.byte, '"test"'), b"test"),
        (pt.TealOp(None, pt.Op.byte, '"\\t\\n\\\\\\""'), b'\t\n\\"'),
        (pt.TealOp(None, pt.Op.byte, "0x"), b""),
        (pt.TealOp(None, pt.Op.byte, "0x00"), b"\x00"),
        (pt.TealOp(None, pt.Op.byte, "0xFF00"), b"\xff\x00"),
        (pt.TealOp(None, pt.Op.byte, "0xff00"), b"\xff\x00"),
        (pt.TealOp(None, pt.Op.byte, "base32()"), b""),
        (pt.TealOp(None, pt.Op.byte, "base32(ORSXG5A)"), b"test"),
        (pt.TealOp(None, pt.Op.byte, "base32(ORSXG5A=)"), b"test"),
        (pt.TealOp(None, pt.Op.byte, "base64()"), b""),
        (pt.TealOp(None, pt.Op.byte, "base64(dGVzdA==)"), b"test"),
        (pt.TealOp(None, pt.Op.byte, "TMPL_NAME"), "TMPL_NAME"),
    ]

    for op, expected in tests:
        actual = extractBytesValue(op)
        assert actual == expected


def test_extractAddrValue():
    tests = [
        (
            pt.TealOp(
                None,
                pt.Op.byte,
                "WSJHNPJ6YCLX5K4GUMQ4ISPK3ABMS3AL3F6CSVQTCUI5F4I65PWEMCWT3M",
            ),
            b"\xb4\x92v\xbd>\xc0\x97~\xab\x86\xa3!\xc4I\xea\xd8\x02\xc9l\x0b\xd9|)V\x13\x15\x11\xd2\xf1\x1e\xeb\xec",
        ),
        (pt.TealOp(None, pt.Op.addr, "TMPL_NAME"), "TMPL_NAME"),
    ]

    for op, expected in tests:
        actual = extractAddrValue(op)
        assert actual == expected


# test case came from: https://gist.github.com/jasonpaulos/99e4f8a75f2fc2ec9b8073c064530359
def test_extractMethodValue():
    tests = [
        (
            pt.TealOp(None, pt.Op.method_signature, '"create(uint64)uint64"'),
            b"\x43\x46\x41\x01",
        ),
        (
            pt.TealOp(None, pt.Op.method_signature, '"update()void"'),
            b"\xa0\xe8\x18\x72",
        ),
        (
            pt.TealOp(None, pt.Op.method_signature, '"optIn(string)string"'),
            b"\xcf\xa6\x8e\x36",
        ),
        (
            pt.TealOp(None, pt.Op.method_signature, '"closeOut()string"'),
            b"\xa9\xf4\x2b\x3d",
        ),
        (
            pt.TealOp(None, pt.Op.method_signature, '"delete()void"'),
            b"\x24\x37\x8d\x3c",
        ),
        (
            pt.TealOp(None, pt.Op.method_signature, '"add(uint64,uint64)uint64"'),
            b"\xfe\x6b\xdf\x69",
        ),
        (pt.TealOp(None, pt.Op.method_signature, '"empty()void"'), b"\xa8\x8c\x26\xa5"),
        (
            pt.TealOp(None, pt.Op.method_signature, '"payment(pay,uint64)bool"'),
            b"\x3e\x3b\x3d\x28",
        ),
        (
            pt.TealOp(
                None,
                pt.Op.method_signature,
                '"referenceTest(account,application,account,asset,account,asset,asset,application,application)uint8[9]"',
            ),
            b"\x0d\xf0\x05\x0f",
        ),
    ]

    for op, expected in tests:
        actual = extractMethodSigValue(op)
        assert actual == expected


def test_createConstantBlocks_empty():
    ops = []

    expected = ops[:]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_no_consts():
    ops = [
        pt.TealOp(None, pt.Op.txn, "Sender"),
        pt.TealOp(None, pt.Op.txn, "Receiver"),
        pt.TealOp(None, pt.Op.eq),
    ]

    expected = ops[:]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_pushint():
    ops = [
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.int, "OptIn"),
        pt.TealOp(None, pt.Op.add),
    ]

    expected = [
        pt.TealOp(None, pt.Op.pushint, 0, "//", 0),
        pt.TealOp(None, pt.Op.pushint, 1, "//", "OptIn"),
        pt.TealOp(None, pt.Op.add),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_intblock_single():
    ops = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, "OptIn"),
        pt.TealOp(None, pt.Op.add),
    ]

    expected = [
        pt.TealOp(None, pt.Op.intcblock, 1),
        pt.TealOp(None, pt.Op.intc_0, "//", 1),
        pt.TealOp(None, pt.Op.intc_0, "//", "OptIn"),
        pt.TealOp(None, pt.Op.add),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_intblock_multiple():
    ops = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, "OptIn"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, "keyreg"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.int, "ClearState"),
        pt.TealOp(None, pt.Op.add),
    ]

    expected = [
        pt.TealOp(None, pt.Op.intcblock, 1, 2, 3),
        pt.TealOp(None, pt.Op.intc_0, "//", 1),
        pt.TealOp(None, pt.Op.intc_0, "//", "OptIn"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.intc_1, "//", 2),
        pt.TealOp(None, pt.Op.intc_1, "//", "keyreg"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.intc_2, "//", 3),
        pt.TealOp(None, pt.Op.intc_2, "//", "ClearState"),
        pt.TealOp(None, pt.Op.add),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_intblock_pushint():
    ops = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, "OptIn"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.int, "ClearState"),
        pt.TealOp(None, pt.Op.add),
    ]

    expected = [
        pt.TealOp(None, pt.Op.intcblock, 3, 1),
        pt.TealOp(None, pt.Op.intc_1, "//", 1),
        pt.TealOp(None, pt.Op.intc_1, "//", "OptIn"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.pushint, 2, "//", 2),
        pt.TealOp(None, pt.Op.intc_0, "//", 3),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.intc_0, "//", 3),
        pt.TealOp(None, pt.Op.intc_0, "//", "ClearState"),
        pt.TealOp(None, pt.Op.add),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_pushbytes():
    ops = [
        pt.TealOp(None, pt.Op.byte, "0x0102"),
        pt.TealOp(None, pt.Op.byte, "0x0103"),
        pt.TealOp(None, pt.Op.method_signature, '"empty()void"'),
        pt.TealOp(None, pt.Op.concat),
    ]

    expected = [
        pt.TealOp(None, pt.Op.pushbytes, "0x0102", "//", "0x0102"),
        pt.TealOp(None, pt.Op.pushbytes, "0x0103", "//", "0x0103"),
        pt.TealOp(None, pt.Op.pushbytes, "0xa88c26a5", "//", '"empty()void"'),
        pt.TealOp(None, pt.Op.concat),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_byteblock_single():
    ops = [
        pt.TealOp(None, pt.Op.byte, "0x0102"),
        pt.TealOp(None, pt.Op.byte, "base64(AQI=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "base32(AEBA====)"),
        pt.TealOp(None, pt.Op.concat),
    ]

    expected = [
        pt.TealOp(None, pt.Op.bytecblock, "0x0102"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "0x0102"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "base64(AQI=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_0, "//", "base32(AEBA====)"),
        pt.TealOp(None, pt.Op.concat),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_byteblock_multiple():
    ops = [
        pt.TealOp(None, pt.Op.byte, "0x0102"),
        pt.TealOp(None, pt.Op.byte, "base64(AQI=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "base32(AEBA====)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, '"test"'),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "base32(ORSXG5A=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(
            None,
            pt.Op.byte,
            "0xb49276bd3ec0977eab86a321c449ead802c96c0bd97c2956131511d2f11eebec",
        ),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(
            None,
            pt.Op.addr,
            "WSJHNPJ6YCLX5K4GUMQ4ISPK3ABMS3AL3F6CSVQTCUI5F4I65PWEMCWT3M",
        ),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.method_signature, '"closeOut()string"'),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "base64(qfQrPQ==)"),
    ]

    expected = [
        pt.TealOp(
            None,
            pt.Op.bytecblock,
            "0x0102",
            "0x74657374",
            "0xb49276bd3ec0977eab86a321c449ead802c96c0bd97c2956131511d2f11eebec",
            "0xa9f42b3d",
        ),
        pt.TealOp(None, pt.Op.bytec_0, "//", "0x0102"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "base64(AQI=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_0, "//", "base32(AEBA====)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_1, "//", '"test"'),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_1, "//", "base32(ORSXG5A=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(
            None,
            pt.Op.bytec_2,
            "//",
            "0xb49276bd3ec0977eab86a321c449ead802c96c0bd97c2956131511d2f11eebec",
        ),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(
            None,
            pt.Op.bytec_2,
            "//",
            "WSJHNPJ6YCLX5K4GUMQ4ISPK3ABMS3AL3F6CSVQTCUI5F4I65PWEMCWT3M",
        ),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_3, "//", '"closeOut()string"'),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_3, "//", "base64(qfQrPQ==)"),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_byteblock_pushbytes():
    ops = [
        pt.TealOp(None, pt.Op.byte, "0x0102"),
        pt.TealOp(None, pt.Op.byte, "base64(AQI=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "base32(AEBA====)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, '"test"'),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "base32(ORSXG5A=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(
            None,
            pt.Op.addr,
            "WSJHNPJ6YCLX5K4GUMQ4ISPK3ABMS3AL3F6CSVQTCUI5F4I65PWEMCWT3M",
        ),
        pt.TealOp(None, pt.Op.concat),
    ]

    expected = [
        pt.TealOp(None, pt.Op.bytecblock, "0x0102", "0x74657374"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "0x0102"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "base64(AQI=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_0, "//", "base32(AEBA====)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_1, "//", '"test"'),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_1, "//", "base32(ORSXG5A=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(
            None,
            pt.Op.pushbytes,
            "0xb49276bd3ec0977eab86a321c449ead802c96c0bd97c2956131511d2f11eebec",
            "//",
            "WSJHNPJ6YCLX5K4GUMQ4ISPK3ABMS3AL3F6CSVQTCUI5F4I65PWEMCWT3M",
        ),
        pt.TealOp(None, pt.Op.concat),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_all():
    ops = [
        pt.TealOp(None, pt.Op.byte, "0x0102"),
        pt.TealOp(None, pt.Op.byte, "base64(AQI=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "base32(AEBA====)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, '"test"'),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "base32(ORSXG5A=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(
            None,
            pt.Op.addr,
            "WSJHNPJ6YCLX5K4GUMQ4ISPK3ABMS3AL3F6CSVQTCUI5F4I65PWEMCWT3M",
        ),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, "OptIn"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.int, "ClearState"),
        pt.TealOp(None, pt.Op.add),
    ]

    expected = [
        pt.TealOp(None, pt.Op.intcblock, 3, 1),
        pt.TealOp(None, pt.Op.bytecblock, "0x0102", "0x74657374"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "0x0102"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "base64(AQI=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_0, "//", "base32(AEBA====)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_1, "//", '"test"'),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_1, "//", "base32(ORSXG5A=)"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(
            None,
            pt.Op.pushbytes,
            "0xb49276bd3ec0977eab86a321c449ead802c96c0bd97c2956131511d2f11eebec",
            "//",
            "WSJHNPJ6YCLX5K4GUMQ4ISPK3ABMS3AL3F6CSVQTCUI5F4I65PWEMCWT3M",
        ),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.intc_1, "//", 1),
        pt.TealOp(None, pt.Op.intc_1, "//", "OptIn"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.pushint, 2, "//", 2),
        pt.TealOp(None, pt.Op.intc_0, "//", 3),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.intc_0, "//", 3),
        pt.TealOp(None, pt.Op.intc_0, "//", "ClearState"),
        pt.TealOp(None, pt.Op.add),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_tmpl_int():
    ops = [
        pt.TealOp(None, pt.Op.int, "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.int, "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.int, "TMPL_INT_2"),
        pt.TealOp(None, pt.Op.add),
    ]

    expected = [
        pt.TealOp(None, pt.Op.intcblock, "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.intc_0, "//", "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.intc_0, "//", "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.pushint, "TMPL_INT_2", "//", "TMPL_INT_2"),
        pt.TealOp(None, pt.Op.add),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_tmpl_int_mixed():
    ops = [
        pt.TealOp(None, pt.Op.int, "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.int, "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.int, "TMPL_INT_2"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
    ]

    expected = [
        pt.TealOp(None, pt.Op.intcblock, "TMPL_INT_1", 0),
        pt.TealOp(None, pt.Op.intc_0, "//", "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.intc_0, "//", "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.pushint, "TMPL_INT_2", "//", "TMPL_INT_2"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.intc_1, "//", 0),
        pt.TealOp(None, pt.Op.intc_1, "//", 0),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.pushint, 1, "//", 1),
        pt.TealOp(None, pt.Op.add),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_tmpl_bytes():
    ops = [
        pt.TealOp(None, pt.Op.byte, "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.byte, "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.byte, "TMPL_BYTES_2"),
        pt.TealOp(None, pt.Op.concat),
    ]

    expected = [
        pt.TealOp(None, pt.Op.bytecblock, "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.pushbytes, "TMPL_BYTES_2", "//", "TMPL_BYTES_2"),
        pt.TealOp(None, pt.Op.concat),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_tmpl_bytes_mixed():
    ops = [
        pt.TealOp(None, pt.Op.byte, "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.byte, "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.byte, "TMPL_BYTES_2"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "0x00"),
        pt.TealOp(None, pt.Op.byte, "0x00"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "0x01"),
        pt.TealOp(None, pt.Op.concat),
    ]

    expected = [
        pt.TealOp(None, pt.Op.bytecblock, "TMPL_BYTES_1", "0x00"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.pushbytes, "TMPL_BYTES_2", "//", "TMPL_BYTES_2"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_1, "//", "0x00"),
        pt.TealOp(None, pt.Op.bytec_1, "//", "0x00"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.pushbytes, "0x01", "//", "0x01"),
        pt.TealOp(None, pt.Op.concat),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_tmpl_all():
    ops = [
        pt.TealOp(None, pt.Op.byte, "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.byte, "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.byte, "TMPL_BYTES_2"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "0x00"),
        pt.TealOp(None, pt.Op.byte, "0x00"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.byte, "0x01"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.len),
        pt.TealOp(None, pt.Op.int, "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.int, "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.int, "TMPL_INT_2"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.eq),
    ]

    expected = [
        pt.TealOp(None, pt.Op.intcblock, "TMPL_INT_1", 0),
        pt.TealOp(None, pt.Op.bytecblock, "TMPL_BYTES_1", "0x00"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.bytec_0, "//", "TMPL_BYTES_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.pushbytes, "TMPL_BYTES_2", "//", "TMPL_BYTES_2"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.bytec_1, "//", "0x00"),
        pt.TealOp(None, pt.Op.bytec_1, "//", "0x00"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.pushbytes, "0x01", "//", "0x01"),
        pt.TealOp(None, pt.Op.concat),
        pt.TealOp(None, pt.Op.len),
        pt.TealOp(None, pt.Op.intc_0, "//", "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.intc_0, "//", "TMPL_INT_1"),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.pushint, "TMPL_INT_2", "//", "TMPL_INT_2"),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.intc_1, "//", 0),
        pt.TealOp(None, pt.Op.intc_1, "//", 0),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.pushint, 1, "//", 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.eq),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_intc():
    """Test scenario where there are more than 4 constants in the intcblock.
    pt.If the 4th constant can't fit in one varuint byte (more than 2**7) it
    should be referenced with the pt.Op.intc 4 command.
    """

    ops = [
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.int, 2**7),
        pt.TealOp(None, pt.Op.int, 2**7),
    ]

    expected = [
        pt.TealOp(None, pt.Op.intcblock, 0, 1, 2, 3, 2**7),
        pt.TealOp(None, pt.Op.intc_0, "//", 0),
        pt.TealOp(None, pt.Op.intc_0, "//", 0),
        pt.TealOp(None, pt.Op.intc_1, "//", 1),
        pt.TealOp(None, pt.Op.intc_1, "//", 1),
        pt.TealOp(None, pt.Op.intc_2, "//", 2),
        pt.TealOp(None, pt.Op.intc_2, "//", 2),
        pt.TealOp(None, pt.Op.intc_3, "//", 3),
        pt.TealOp(None, pt.Op.intc_3, "//", 3),
        pt.TealOp(None, pt.Op.intc, 4, "//", 2**7),
        pt.TealOp(None, pt.Op.intc, 4, "//", 2**7),
    ]

    actual = createConstantBlocks(ops)
    assert actual == expected


def test_createConstantBlocks_small_constant():
    """pt.If a constant cannot be referenced using the intc_[0..3] commands
    and it can be stored in one varuint it byte then pt.Op.pushint is used.
    """

    for cur in range(4, 2**7):
        ops = [
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.int, cur),
            pt.TealOp(None, pt.Op.int, cur),
        ]

        expected = [
            pt.TealOp(None, pt.Op.intcblock, 0, 1, 2, 3),
            pt.TealOp(None, pt.Op.intc_0, "//", 0),
            pt.TealOp(None, pt.Op.intc_0, "//", 0),
            pt.TealOp(None, pt.Op.intc_1, "//", 1),
            pt.TealOp(None, pt.Op.intc_1, "//", 1),
            pt.TealOp(None, pt.Op.intc_2, "//", 2),
            pt.TealOp(None, pt.Op.intc_2, "//", 2),
            pt.TealOp(None, pt.Op.intc_3, "//", 3),
            pt.TealOp(None, pt.Op.intc_3, "//", 3),
            pt.TealOp(None, pt.Op.pushint, cur, "//", cur),
            pt.TealOp(None, pt.Op.pushint, cur, "//", cur),
        ]

        actual = createConstantBlocks(ops)
        assert actual == expected



import pytest

import pyteal as pt
from pyteal import abi

options = pt.CompileOptions(version=5)


def test_ReferenceTypeSpecs_list():
    assert abi.ReferenceTypeSpecs == [
        abi.AccountTypeSpec(),
        abi.AssetTypeSpec(),
        abi.ApplicationTypeSpec(),
    ]


def test_ReferenceType_referenced_index():
    for value in (abi.Account(), abi.Asset(), abi.Application()):
        expr = value.referenced_index()
        assert expr.type_of() == pt.TealType.uint64
        assert expr.has_return() is False

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(expr, pt.Op.load, value.stored_value.slot),
            ]
        )
        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected


def test_ReferenceType_encode():
    for value in (abi.Account(), abi.Asset(), abi.Application()):
        with pytest.raises(
            pt.TealInputError, match=r"A ReferenceType cannot be encoded$"
        ):
            value.encode()


def test_ReferenceType_decode():
    encoded = pt.Bytes("encoded")
    for value in (abi.Account(), abi.Asset(), abi.Application()):
        for start_index in (None, pt.Int(1)):
            for end_index in (None, pt.Int(2)):
                for length in (None, pt.Int(3)):
                    expr = value.decode(
                        encoded,
                        start_index=start_index,
                        end_index=end_index,
                        length=length,
                    )
                    assert expr.type_of() == pt.TealType.none
                    assert expr.has_return() is False

                    expected_decoding = value.stored_value.store(
                        pt.GetByte(
                            encoded,
                            start_index if start_index is not None else pt.Int(0),
                        )
                    )
                    expected, _ = expected_decoding.__teal__(options)
                    expected.addIncoming()
                    expected = pt.TealBlock.NormalizeBlocks(expected)

                    actual, _ = expr.__teal__(options)
                    actual.addIncoming()
                    actual = pt.TealBlock.NormalizeBlocks(actual)

                    with pt.TealComponent.Context.ignoreExprEquality():
                        assert actual == expected


def test_Account_str():
    assert str(abi.AccountTypeSpec()) == "account"


def test_AccountTypeSpec_is_dynamic():
    assert not (abi.AccountTypeSpec()).is_dynamic()


def test_AccountTypeSpec_new_instance():
    assert isinstance(abi.AccountTypeSpec().new_instance(), abi.Account)


def test_AccountTypeSpec_eq():
    assert abi.AccountTypeSpec() == abi.AccountTypeSpec()

    for otherType in (
        abi.ByteTypeSpec(),
        abi.Uint8TypeSpec(),
        abi.AddressTypeSpec(),
    ):
        assert abi.AccountTypeSpec() != otherType


def test_Account_typespec():
    assert abi.Account().type_spec() == abi.AccountTypeSpec()


def test_Account_address():
    value = abi.Account()
    expr = value.address()
    assert expr.type_of() == pt.TealType.bytes
    assert expr.has_return() is False

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.load, value.stored_value.slot),
            pt.TealOp(None, pt.Op.txnas, "Accounts"),
        ]
    )
    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_Account_params():
    value = abi.Account()

    params = value.params()

    assert type(params) is pt.AccountParamObject

    expected = value.referenced_index()
    actual = params._account

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual.__teal__(options) == expected.__teal__(options)


def test_Account_asset_holding():
    value = abi.Account()

    assets = ((pt.Int(6), pt.Int(6)), (a := abi.Asset(), a.referenced_index()))

    for asset, expected_asset in assets:
        holding = value.asset_holding(asset)

        assert type(holding) is pt.AssetHoldingObject

        expected_account = value.referenced_index()
        actual_account = holding._account

        actual_asset = holding._asset

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual_account.__teal__(options) == expected_account.__teal__(
                options
            )
            assert actual_asset.__teal__(options) == expected_asset.__teal__(options)


def test_Asset_str():
    assert str(abi.AssetTypeSpec()) == "asset"


def test_AssetTypeSpec_is_dynamic():
    assert not (abi.AssetTypeSpec()).is_dynamic()


def test_AssetTypeSpec_new_instance():
    assert isinstance(abi.AssetTypeSpec().new_instance(), abi.Asset)


def test_AssetTypeSpec_eq():
    assert abi.AssetTypeSpec() == abi.AssetTypeSpec()

    for otherType in (
        abi.ByteTypeSpec(),
        abi.Uint8TypeSpec(),
        abi.AddressTypeSpec(),
    ):
        assert abi.AssetTypeSpec() != otherType


def test_Asset_typespec():
    assert abi.Asset().type_spec() == abi.AssetTypeSpec()


def test_Asset_asset_id():
    value = abi.Asset()
    expr = value.asset_id()
    assert expr.type_of() == pt.TealType.uint64
    assert expr.has_return() is False

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.load, value.stored_value.slot),
            pt.TealOp(None, pt.Op.txnas, "Assets"),
        ]
    )
    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_Asset_holding():
    value = abi.Asset()

    accounts = (
        (pt.Int(6), pt.Int(6)),
        (
            pt.Addr("QSA6K5MNJPEGO5SDSWXBM3K4UEI3Q2NCPS2OUXVJI5QPCHMVI27MFRSHKI"),
            pt.Addr("QSA6K5MNJPEGO5SDSWXBM3K4UEI3Q2NCPS2OUXVJI5QPCHMVI27MFRSHKI"),
        ),
        (a := abi.Account(), a.referenced_index()),
    )

    for account, expected_account in accounts:
        holding = value.holding(account)

        assert type(holding) is pt.AssetHoldingObject

        expected_asset = value.referenced_index()
        actual_asset = holding._asset

        actual_account = holding._account

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual_asset.__teal__(options) == expected_asset.__teal__(options)
            assert actual_account.__teal__(options) == expected_account.__teal__(
                options
            )


def test_Asset_params():
    value = abi.Asset()

    params = value.params()

    assert type(params) is pt.AssetParamObject

    expected = value.referenced_index()
    actual = params._asset

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual.__teal__(options) == expected.__teal__(options)


def test_Application_str():
    assert str(abi.ApplicationTypeSpec()) == "application"


def test_ApplicationTypeSpec_is_dynamic():
    assert not (abi.ApplicationTypeSpec()).is_dynamic()


def test_ApplicationTypeSpec_new_instance():
    assert isinstance(abi.ApplicationTypeSpec().new_instance(), abi.Application)


def test_ApplicationTypeSpec_eq():
    assert abi.ApplicationTypeSpec() == abi.ApplicationTypeSpec()

    for otherType in (
        abi.ByteTypeSpec(),
        abi.Uint8TypeSpec(),
        abi.AddressTypeSpec(),
    ):
        assert abi.ApplicationTypeSpec() != otherType


def test_Application_typespec():
    assert abi.Application().type_spec() == abi.ApplicationTypeSpec()


def test_Application_application_id():
    value = abi.Application()
    expr = value.application_id()
    assert expr.type_of() == pt.TealType.uint64
    assert expr.has_return() is False

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.load, value.stored_value.slot),
            pt.TealOp(None, pt.Op.txnas, "Applications"),
        ]
    )
    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_Application_params():
    value = abi.Application()

    params = value.params()

    assert type(params) is pt.AppParamObject

    expected = value.referenced_index()
    actual = params._app

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual.__teal__(options) == expected.__teal__(options)



from typing import TYPE_CHECKING, Tuple
from pyteal.ast.abi import BaseType
from pyteal.types import TealType
from pyteal.errors import TealInputError
from pyteal.ast.expr import Expr
from pyteal.ast.unaryexpr import Log
from pyteal.ast.naryexpr import Concat
from pyteal.ast.bytes import Bytes
from pyteal.ir import TealBlock, TealSimpleBlock, Op
from pyteal.config import RETURN_HASH_PREFIX

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class MethodReturn(Expr):
    def __init__(self, arg: BaseType):
        super().__init__()
        if not isinstance(arg, BaseType):
            raise TealInputError(f"Expecting an ABI type argument but get {arg}")
        self.arg = arg

    def __teal__(self, options: "CompileOptions") -> Tuple[TealBlock, TealSimpleBlock]:
        if options.version < Op.log.min_version:
            raise TealInputError(
                f"current version {options.version} is lower than log's min version {Op.log.min_version}"
            )
        return Log(Concat(Bytes(RETURN_HASH_PREFIX), self.arg.encode())).__teal__(
            options
        )

    def __str__(self) -> str:
        return f"(MethodReturn {self.arg.type_spec()})"

    def type_of(self) -> TealType:
        return TealType.none

    def has_return(self) -> bool:
        return False


MethodReturn.__module__ = "pyteal.abi"



from itertools import product
from typing import List, Literal, Optional, cast

import pytest
from dataclasses import dataclass

import pyteal as pt
from pyteal.ast.subroutine import ABIReturnSubroutine, evaluate_subroutine

options = pt.CompileOptions(version=5)


def test_subroutine_definition():
    def fn0Args():
        return pt.Return()

    def fn1Args(a1):
        return pt.Return()

    def fn2Args(a1, a2):
        return pt.Return()

    def fn10Args(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10):
        return pt.Return()

    lam0Args = lambda: pt.Return()  # noqa: E731
    lam1Args = lambda a1: pt.Return()  # noqa: E731
    lam2Args = lambda a1, a2: pt.Return()  # noqa: E731
    lam10Args = (
        lambda a1, a2, a3, a4, a5, a6, a7, a8, a9, a10: pt.Return()
    )  # noqa: E731

    def fnWithExprAnnotations(a: pt.Expr, b: pt.Expr) -> pt.Expr:
        return pt.Return()

    def fnWithOnlyReturnExprAnnotations(a, b) -> pt.Expr:
        return pt.Return()

    def fnWithOnlyArgExprAnnotations(a: pt.Expr, b: pt.Expr):
        return pt.Return()

    def fnWithPartialExprAnnotations(a, b: pt.Expr) -> pt.Expr:
        return pt.Return()

    cases = (
        (fn0Args, 0, "fn0Args"),
        (fn1Args, 1, "fn1Args"),
        (fn2Args, 2, "fn2Args"),
        (fn10Args, 10, "fn10Args"),
        (lam0Args, 0, "<lambda>"),
        (lam1Args, 1, "<lambda>"),
        (lam2Args, 2, "<lambda>"),
        (lam10Args, 10, "<lambda>"),
        (fnWithExprAnnotations, 2, "fnWithExprAnnotations"),
        (fnWithOnlyReturnExprAnnotations, 2, "fnWithOnlyReturnExprAnnotations"),
        (fnWithOnlyArgExprAnnotations, 2, "fnWithOnlyArgExprAnnotations"),
        (fnWithPartialExprAnnotations, 2, "fnWithPartialExprAnnotations"),
    )

    for (fn, numArgs, name) in cases:
        definition = pt.SubroutineDefinition(fn, pt.TealType.none)
        assert definition.argument_count() == numArgs
        assert definition.name() == name

        if numArgs > 0:
            with pytest.raises(pt.TealInputError):
                definition.invoke([pt.Int(1)] * (numArgs - 1))

        with pytest.raises(pt.TealInputError):
            definition.invoke([pt.Int(1)] * (numArgs + 1))

        if numArgs > 0:
            with pytest.raises(pt.TealInputError):
                definition.invoke([1] * numArgs)

        args = [pt.Int(1)] * numArgs
        invocation = definition.invoke(args)
        assert isinstance(invocation, pt.SubroutineCall)
        assert invocation.subroutine is definition
        assert invocation.args == args


@dataclass
class ABISubroutineTC:
    definition: pt.ABIReturnSubroutine
    arg_instances: list[pt.Expr | pt.abi.BaseType]
    name: str
    ret_type: str | pt.abi.TypeSpec
    signature: Optional[str]


def test_abi_subroutine_definition():
    @pt.ABIReturnSubroutine
    def fn_0arg_0ret() -> pt.Expr:
        return pt.Return()

    @pt.ABIReturnSubroutine
    def fn_0arg_uint64_ret(*, output: pt.abi.Uint64) -> pt.Expr:
        return output.set(1)

    @pt.ABIReturnSubroutine
    def fn_1arg_0ret(a: pt.abi.Uint64) -> pt.Expr:
        return pt.Return()

    @pt.ABIReturnSubroutine
    def fn_1arg_1ret(a: pt.abi.Uint64, *, output: pt.abi.Uint64) -> pt.Expr:
        return output.set(a)

    @pt.ABIReturnSubroutine
    def fn_2arg_0ret(
        a: pt.abi.Uint64, b: pt.abi.StaticArray[pt.abi.Byte, Literal[10]]
    ) -> pt.Expr:
        return pt.Return()

    @pt.ABIReturnSubroutine
    def fn_2arg_1ret(
        a: pt.abi.Uint64,
        b: pt.abi.StaticArray[pt.abi.Byte, Literal[10]],
        *,
        output: pt.abi.Byte,
    ) -> pt.Expr:
        return output.set(b[a.get() % pt.Int(10)])

    @pt.ABIReturnSubroutine
    def fn_2arg_1ret_with_expr(
        a: pt.Expr,
        b: pt.abi.StaticArray[pt.abi.Byte, Literal[10]],
        *,
        output: pt.abi.Byte,
    ) -> pt.Expr:
        return output.set(b[a % pt.Int(10)])

    @pt.ABIReturnSubroutine
    def fn_w_tuple1arg(
        a: pt.Expr,
        b: pt.abi.Tuple1[pt.abi.Byte],
        *,
        output: pt.abi.Byte,
    ) -> pt.Expr:
        return output.set(pt.Int(1))

    cases = (
        ABISubroutineTC(fn_0arg_0ret, [], "fn_0arg_0ret", "void", "fn_0arg_0ret()void"),
        ABISubroutineTC(
            fn_0arg_uint64_ret,
            [],
            "fn_0arg_uint64_ret",
            pt.abi.Uint64TypeSpec(),
            "fn_0arg_uint64_ret()uint64",
        ),
        ABISubroutineTC(
            fn_1arg_0ret,
            [pt.abi.Uint64()],
            "fn_1arg_0ret",
            "void",
            "fn_1arg_0ret(uint64)void",
        ),
        ABISubroutineTC(
            fn_1arg_1ret,
            [pt.abi.Uint64()],
            "fn_1arg_1ret",
            pt.abi.Uint64TypeSpec(),
            "fn_1arg_1ret(uint64)uint64",
        ),
        ABISubroutineTC(
            fn_2arg_0ret,
            [
                pt.abi.Uint64(),
                pt.abi.StaticArray(
                    pt.abi.StaticArrayTypeSpec(pt.abi.ByteTypeSpec(), 10)
                ),
            ],
            "fn_2arg_0ret",
            "void",
            "fn_2arg_0ret(uint64,byte[10])void",
        ),
        ABISubroutineTC(
            fn_2arg_1ret,
            [
                pt.abi.Uint64(),
                pt.abi.StaticArray(
                    pt.abi.StaticArrayTypeSpec(pt.abi.ByteTypeSpec(), 10)
                ),
            ],
            "fn_2arg_1ret",
            pt.abi.ByteTypeSpec(),
            "fn_2arg_1ret(uint64,byte[10])byte",
        ),
        ABISubroutineTC(
            fn_2arg_1ret_with_expr,
            [
                pt.Int(5),
                pt.abi.StaticArray(
                    pt.abi.StaticArrayTypeSpec(pt.abi.ByteTypeSpec(), 10)
                ),
            ],
            "fn_2arg_1ret_with_expr",
            pt.abi.ByteTypeSpec(),
            None,
        ),
        ABISubroutineTC(
            fn_w_tuple1arg,
            [
                pt.Int(5),
                pt.abi.make(pt.abi.Tuple1[pt.abi.Byte]),
            ],
            "fn_w_tuple1arg",
            pt.abi.ByteTypeSpec(),
            None,
        ),
    )

    for case in cases:
        assert case.definition.subroutine.argument_count() == len(case.arg_instances)
        assert case.definition.name() == case.name

        if len(case.arg_instances) > 0:
            with pytest.raises(pt.TealInputError):
                case.definition(*case.arg_instances[:-1])

        with pytest.raises(pt.TealInputError):
            case.definition(*(case.arg_instances + [pt.abi.Uint64()]))

        assert case.definition.type_of() == case.ret_type
        invoked = case.definition(*case.arg_instances)
        assert isinstance(
            invoked, (pt.Expr if case.ret_type == "void" else pt.abi.ReturnedValue)
        )
        assert case.definition.is_abi_routable() == all(
            map(lambda x: isinstance(x, pt.abi.BaseType), case.arg_instances)
        )

        if case.definition.is_abi_routable():
            assert case.definition.method_signature() == cast(str, case.signature)
        else:
            with pytest.raises(pt.TealInputError):
                case.definition.method_signature()


def test_subroutine_return_reference():
    @ABIReturnSubroutine
    def invalid_ret_type(*, output: pt.abi.Account):
        return output.decode(pt.Bytes(b"\x00"))

    with pytest.raises(pt.TealInputError):
        invalid_ret_type.method_signature()

    @ABIReturnSubroutine
    def invalid_ret_type_collection(
        *, output: pt.abi.Tuple2[pt.abi.Account, pt.abi.Uint64]
    ):
        return output.set(pt.abi.Account(), pt.abi.Uint64())

    with pytest.raises(pt.TealInputError):
        invalid_ret_type_collection.method_signature()

    @ABIReturnSubroutine
    def invalid_ret_type_collection_nested(
        *, output: pt.abi.DynamicArray[pt.abi.Tuple2[pt.abi.Account, pt.abi.Uint64]]
    ):
        return output.set(
            pt.abi.make(
                pt.abi.DynamicArray[pt.abi.Tuple2[pt.abi.Account, pt.abi.Uint64]]
            )
        )

    with pytest.raises(pt.TealInputError):
        invalid_ret_type_collection_nested.method_signature()


def test_subroutine_definition_validate():
    """
    DFS through SubroutineDefinition.validate()'s logic
    """

    def mock_subroutine_definition(implementation, has_abi_output=False):
        mock = pt.SubroutineDefinition(lambda: pt.Return(pt.Int(1)), pt.TealType.uint64)
        mock._validate()  # haven't failed with dummy implementation
        mock.implementation = implementation
        mock.has_abi_output = has_abi_output
        return mock

    not_callable = mock_subroutine_definition("I'm not callable")
    with pytest.raises(pt.TealInputError) as tie:
        not_callable._validate()

    assert tie.value == pt.TealInputError(
        "Input to SubroutineDefinition is not callable"
    )

    # input_types:

    three_params = mock_subroutine_definition(lambda x, y, z: pt.Return(pt.Int(1)))

    params, anns, arg_types, byrefs, abi_args, output_kwarg = three_params._validate()
    assert len(params) == 3
    assert anns == {}
    assert all(at is pt.Expr for at in arg_types)
    assert byrefs == set()
    assert abi_args == {}
    assert output_kwarg == {}

    def bad_return_impl() -> str:
        return pt.Return(pt.Int(1))  # type: ignore

    bad_return = mock_subroutine_definition(bad_return_impl)
    with pytest.raises(pt.TealInputError) as tie:
        bad_return._validate()

    assert tie.value == pt.TealInputError(
        "Function has return of disallowed type <class 'str'>. Only Expr is allowed"
    )

    # now we iterate through the implementation params validating each as we go

    def var_abi_output_impl(*, output: pt.abi.Uint16):
        pt.Return(pt.Int(1))  # this is wrong but ignored

    # raises without abi_output_arg_name:
    var_abi_output_noname = mock_subroutine_definition(var_abi_output_impl)
    with pytest.raises(pt.TealInputError) as tie:
        var_abi_output_noname._validate()

    assert tie.value == pt.TealInputError(
        "Function has a parameter type that is not allowed in a subroutine: parameter output with type KEYWORD_ONLY"
    )

    # copacetic abi output:
    var_abi_output = mock_subroutine_definition(
        var_abi_output_impl, has_abi_output=True
    )
    params, anns, arg_types, byrefs, abi_args, output_kwarg = var_abi_output._validate()
    assert len(params) == 1
    assert anns == {"output": pt.abi.Uint16}
    assert all(at is pt.Expr for at in arg_types)
    assert byrefs == set()
    assert abi_args == {}
    assert output_kwarg == {"output": pt.abi.Uint16TypeSpec()}

    var_positional = mock_subroutine_definition(lambda *args: pt.Return(pt.Int(1)))
    with pytest.raises(pt.TealInputError) as tie:
        var_positional._validate()

    assert tie.value == pt.TealInputError(
        "Function has a parameter type that is not allowed in a subroutine: parameter args with type VAR_POSITIONAL"
    )

    kw_only = mock_subroutine_definition(lambda *, kw: pt.Return(pt.Int(1)))
    with pytest.raises(pt.TealInputError) as tie:
        kw_only._validate()

    assert tie.value == pt.TealInputError(
        "Function has a parameter type that is not allowed in a subroutine: parameter kw with type KEYWORD_ONLY"
    )

    var_keyword = mock_subroutine_definition(lambda **kw: pt.Return(pt.Int(1)))
    with pytest.raises(pt.TealInputError) as tie:
        var_keyword._validate()

    assert tie.value == pt.TealInputError(
        "Function has a parameter type that is not allowed in a subroutine: parameter kw with type VAR_KEYWORD"
    )

    param_default = mock_subroutine_definition(lambda x="niiiice": pt.Return(pt.Int(1)))
    with pytest.raises(pt.TealInputError) as tie:
        param_default._validate()

    assert tie.value == pt.TealInputError(
        "Function has a parameter with a default value, which is not allowed in a subroutine: x"
    )

    # Now we get to _validate_annotation():
    one_vanilla = mock_subroutine_definition(lambda x: pt.Return(pt.Int(1)))

    params, anns, arg_types, byrefs, abi_args, output_kwarg = one_vanilla._validate()
    assert len(params) == 1
    assert anns == {}
    assert all(at is pt.Expr for at in arg_types)
    assert byrefs == set()
    assert abi_args == {}
    assert output_kwarg == {}

    def one_expr_impl(x: pt.Expr):
        return pt.Return(pt.Int(1))

    one_expr = mock_subroutine_definition(one_expr_impl)
    params, anns, arg_types, byrefs, abi_args, output_kwarg = one_expr._validate()
    assert len(params) == 1
    assert anns == {"x": pt.Expr}
    assert all(at is pt.Expr for at in arg_types)
    assert byrefs == set()
    assert abi_args == {}
    assert output_kwarg == {}

    def one_scratchvar_impl(x: pt.ScratchVar):
        return pt.Return(pt.Int(1))

    one_scratchvar = mock_subroutine_definition(one_scratchvar_impl)
    params, anns, arg_types, byrefs, abi_args, output_kwarg = one_scratchvar._validate()
    assert len(params) == 1
    assert anns == {"x": pt.ScratchVar}
    assert all(at is pt.ScratchVar for at in arg_types)
    assert byrefs == {"x"}
    assert abi_args == {}
    assert output_kwarg == {}

    # not is_class()
    def one_nontype_impl(x: "blahBlah"):  # type: ignore # noqa: F821
        return pt.Return(pt.Int(1))

    one_nontype = mock_subroutine_definition(one_nontype_impl)
    with pytest.raises(pt.TealInputError) as tie:
        one_nontype._validate()

    assert tie.value == pt.TealInputError(
        "Function has parameter x of declared type blahBlah which is not a class"
    )

    def one_dynscratchvar_impl(x: pt.DynamicScratchVar):
        return pt.Return(pt.Int(1))

    one_dynscratchvar = mock_subroutine_definition(one_dynscratchvar_impl)
    with pytest.raises(pt.TealInputError) as tie:
        one_dynscratchvar._validate()

    assert tie.value == pt.TealInputError(
        "Function has parameter x of disallowed type <class 'pyteal.DynamicScratchVar'>. Only the types (<class 'pyteal.Expr'>, <class 'pyteal.ScratchVar'>, 'ABI') are allowed"
    )

    # Now we're back to _validate() main body and looking at input_types

    three_params_with_output = mock_subroutine_definition(
        lambda x, y, z, *, output: pt.Return(pt.Int(1)), has_abi_output=True
    )
    four_inputs = [
        pt.TealType.uint64,
        pt.TealType.uint64,
        pt.TealType.bytes,
        pt.TealType.uint64,
    ]

    two_inputs = [pt.TealType.uint64, pt.TealType.bytes]
    with pytest.raises(pt.TealInputError) as tie:
        three_params._validate(input_types=two_inputs)

    assert tie.value == pt.TealInputError(
        "Provided number of input_types (2) does not match detected number of input parameters (3)"
    )

    three_inputs_with_a_wrong_type = [pt.TealType.uint64, pt.Expr, pt.TealType.bytes]

    with pytest.raises(pt.TealInputError) as tie:
        three_params._validate(input_types=three_inputs_with_a_wrong_type)

    assert tie.value == pt.TealInputError(
        "Function has input type <class 'pyteal.Expr'> for parameter y which is not a TealType"
    )

    with pytest.raises(pt.TealInputError) as tie:
        three_params._validate(
            input_types=[pt.TealType.uint64, pt.Expr, pt.TealType.anytype]
        )

    assert tie.value == pt.TealInputError(
        "Function has input type <class 'pyteal.Expr'> for parameter y which is not a TealType"
    )

    with pytest.raises(pt.TealInputError) as tie:
        three_params._validate(
            input_types=[pt.TealType.uint64, None, pt.TealType.anytype]
        )
    assert tie.value == pt.TealInputError(
        "input_type for y is unspecified i.e. None but this is only allowed for ABI arguments"
    )

    # this one gets caught inside of _validate_annotation()
    with pytest.raises(pt.TealInputError) as tie:
        three_params_with_output._validate(input_types=four_inputs)

    # everything should be copacetic
    for x, y, z in product(pt.TealType, pt.TealType, pt.TealType):
        (
            params,
            anns,
            arg_types,
            byrefs,
            abi_args,
            output_kwarg,
        ) = three_params._validate(input_types=[x, y, z])
        assert len(params) == 3
        assert anns == {}
        assert all(at is pt.Expr for at in arg_types)
        assert byrefs == set()
        assert abi_args == {}
        assert output_kwarg == {}

    # annotation / abi type handling:
    abi_annotation_examples = {
        pt.abi.Address: pt.abi.AddressTypeSpec(),
        pt.abi.Bool: pt.abi.BoolTypeSpec(),
        pt.abi.Byte: pt.abi.ByteTypeSpec(),
        pt.abi.DynamicArray[pt.abi.Bool]: pt.abi.DynamicArrayTypeSpec(
            pt.abi.BoolTypeSpec()
        ),
        pt.abi.StaticArray[pt.abi.Uint32, Literal[10]]: pt.abi.StaticArrayTypeSpec(
            pt.abi.Uint32TypeSpec(), 10
        ),
        pt.abi.String: pt.abi.StringTypeSpec(),
        pt.abi.Tuple2[pt.abi.Bool, pt.abi.Uint32]: pt.abi.TupleTypeSpec(
            pt.abi.BoolTypeSpec(), pt.abi.Uint32TypeSpec()
        ),
        pt.abi.Uint8: pt.abi.Uint8TypeSpec(),
        pt.abi.Uint16: pt.abi.Uint16TypeSpec(),
        pt.abi.Uint32: pt.abi.Uint32TypeSpec(),
        pt.abi.Uint64: pt.abi.Uint64TypeSpec(),
    }

    anns = (pt.Expr, pt.ScratchVar) + tuple(abi_annotation_examples.keys())
    for x_ann, z_ann in product(anns, anns):

        def mocker_impl(x: x_ann, y, z: z_ann):
            return pt.Return(pt.Int(1))

        mocker = mock_subroutine_definition(mocker_impl)
        params, anns, arg_types, byrefs, abis, output_kwarg = mocker._validate()
        print(
            f"{x_ann=}, {z_ann=}, {params=}, {anns=}, {arg_types=}, {byrefs=}, {abis=}, {output_kwarg=}"
        )

        assert len(params) == 3

        assert anns == {"x": x_ann, "z": z_ann}

        assert (
            (arg_types[0] is x_ann or arg_types[0] == abi_annotation_examples[x_ann])
            and arg_types[1] is pt.Expr
            and (
                arg_types[2] is z_ann or arg_types[2] == abi_annotation_examples[z_ann]
            )
        ), f"{arg_types[0]} -> {x_ann} and {arg_types[1]} -> {pt.Expr} and {arg_types[2]} -> {z_ann}"

        assert byrefs == set(["x"] if x_ann is pt.ScratchVar else []) | set(
            ["z"] if z_ann is pt.ScratchVar else []
        )
        expected_abis = {}
        if x_ann not in (pt.Expr, pt.ScratchVar):
            expected_abis["x"] = abi_annotation_examples[x_ann]
        if z_ann not in (pt.Expr, pt.ScratchVar):
            expected_abis["z"] = abi_annotation_examples[z_ann]
        assert abis == expected_abis


def test_subroutine_invocation_param_types():
    def fnWithNoAnnotations(a, b):
        return pt.Return()

    def fnWithExprAnnotations(a: pt.Expr, b: pt.Expr) -> pt.Expr:
        return pt.Return()

    def fnWithSVAnnotations(a: pt.ScratchVar, b: pt.ScratchVar):
        return pt.Return()

    def fnWithABIAnnotations(
        a: pt.abi.Byte,
        b: pt.abi.StaticArray[pt.abi.Uint32, Literal[10]],
        c: pt.abi.DynamicArray[pt.abi.Bool],
    ):
        return pt.Return()

    def fnWithMixedAnns1(a: pt.ScratchVar, b: pt.Expr) -> pt.Expr:
        return pt.Return()

    def fnWithMixedAnns2(a: pt.ScratchVar, b) -> pt.Expr:
        return pt.Return()

    def fnWithMixedAnns3(a: pt.Expr, b: pt.ScratchVar):
        return pt.Return()

    def fnWithMixedAnns4(a: pt.ScratchVar, b, c: pt.abi.Uint16) -> pt.Expr:
        return pt.Return()

    sv = pt.ScratchVar()
    x = pt.Int(42)
    s = pt.Bytes("hello")
    av_u16 = pt.abi.Uint16()
    av_bool_dym_arr = pt.abi.DynamicArray(
        pt.abi.DynamicArrayTypeSpec(pt.abi.BoolTypeSpec())
    )
    av_u32_static_arr = pt.abi.StaticArray(
        pt.abi.StaticArrayTypeSpec(pt.abi.Uint32TypeSpec(), 10)
    )
    av_bool = pt.abi.Bool()
    av_byte = pt.abi.Byte()

    cases = [
        ("vanilla 1", fnWithNoAnnotations, [x, s], None),
        ("vanilla 2", fnWithNoAnnotations, [x, x], None),
        ("vanilla no sv's allowed 1", fnWithNoAnnotations, [x, sv], pt.TealInputError),
        ("exprs 1", fnWithExprAnnotations, [x, s], None),
        ("exprs 2", fnWithExprAnnotations, [x, x], None),
        ("exprs no sv's allowed 1", fnWithExprAnnotations, [x, sv], pt.TealInputError),
        ("all sv's 1", fnWithSVAnnotations, [sv, sv], None),
        ("all sv's but strings", fnWithSVAnnotations, [s, s], pt.TealInputError),
        ("all sv's but ints", fnWithSVAnnotations, [x, x], pt.TealInputError),
        (
            "all abi's 1",
            fnWithABIAnnotations,
            [av_byte, av_u32_static_arr, av_bool_dym_arr],
            None,
        ),
        (
            "all abi's but ints 1",
            fnWithABIAnnotations,
            [x, av_u32_static_arr, av_bool_dym_arr],
            pt.TealInputError,
        ),
        (
            "all abi's but ints 2",
            fnWithABIAnnotations,
            [x, av_u32_static_arr, x],
            pt.TealInputError,
        ),
        ("all abi's but ints 3", fnWithABIAnnotations, [x, x, x], pt.TealInputError),
        (
            "all abi's but sv's 1",
            fnWithABIAnnotations,
            [sv, av_u32_static_arr, av_bool_dym_arr],
            pt.TealInputError,
        ),
        (
            "all abi's but sv's 2",
            fnWithABIAnnotations,
            [av_byte, av_u32_static_arr, sv],
            pt.TealInputError,
        ),
        (
            "all abi's but sv's 3",
            fnWithABIAnnotations,
            [av_byte, sv, av_u32_static_arr],
            pt.TealInputError,
        ),
        (
            "all abi's but wrong typed 1",
            fnWithABIAnnotations,
            [av_u32_static_arr, av_u32_static_arr, av_bool_dym_arr],
            pt.TealInputError,
        ),
        (
            "all abi's but wrong typed 2",
            fnWithABIAnnotations,
            [av_bool, av_bool_dym_arr, av_u16],
            pt.TealInputError,
        ),
        (
            "all abi's but wrong typed 3",
            fnWithABIAnnotations,
            [av_u16, av_bool, av_byte],
            pt.TealInputError,
        ),
        ("mixed1 copacetic", fnWithMixedAnns1, [sv, x], None),
        ("mixed1 flipped", fnWithMixedAnns1, [x, sv], pt.TealInputError),
        ("mixed1 missing the sv", fnWithMixedAnns1, [x, s], pt.TealInputError),
        ("mixed1 missing the non-sv", fnWithMixedAnns1, [sv, sv], pt.TealInputError),
        ("mixed2 copacetic", fnWithMixedAnns2, [sv, x], None),
        ("mixed2 flipped", fnWithMixedAnns2, [x, sv], pt.TealInputError),
        ("mixed2 missing the sv", fnWithMixedAnns2, [x, s], pt.TealInputError),
        ("mixed2 missing the non-sv", fnWithMixedAnns2, [sv, sv], pt.TealInputError),
        ("mixed3 copacetic", fnWithMixedAnns3, [s, sv], None),
        ("mixed3 flipped", fnWithMixedAnns3, [sv, x], pt.TealInputError),
        ("mixed3 missing the sv", fnWithMixedAnns3, [x, s], pt.TealInputError),
        ("mixed anno", fnWithMixedAnns4, [sv, x, av_u16], None),
        (
            "mixed anno but wrong typed 1",
            fnWithMixedAnns4,
            [av_byte, x, av_u16],
            pt.TealInputError,
        ),
        (
            "mixed anno but wrong typed 2",
            fnWithMixedAnns4,
            [sv, av_byte, sv],
            pt.TealInputError,
        ),
        (
            "mixed anno but wrong typed 3",
            fnWithMixedAnns4,
            [sv, x, av_byte],
            pt.TealInputError,
        ),
    ]
    for case_name, fn, args, err in cases:
        definition = pt.SubroutineDefinition(fn, pt.TealType.none)
        assert definition.argument_count() == len(args), case_name
        assert definition.name() == fn.__name__, case_name

        if err is None:
            assert len(definition.by_ref_args) == len(
                [x for x in args if isinstance(x, pt.ScratchVar)]
            ), case_name

            invocation = definition.invoke(args)
            assert isinstance(invocation, pt.SubroutineCall), case_name
            assert invocation.subroutine is definition, case_name
            assert invocation.args == args, case_name
            assert invocation.has_return() is False, case_name

        else:
            try:
                with pytest.raises(err):
                    definition.invoke(args)
            except Exception as e:
                assert (
                    not e
                ), f"EXPECTED ERROR of type {err}. encountered unexpected error during invocation case <{case_name}>: {e}"


def test_abi_subroutine_calling_param_types():
    @pt.ABIReturnSubroutine
    def fn_log_add(a: pt.abi.Uint64, b: pt.abi.Uint32) -> pt.Expr:
        return pt.Seq(pt.Log(pt.Itob(a.get() + b.get())), pt.Return())

    @pt.ABIReturnSubroutine
    def fn_ret_add(
        a: pt.abi.Uint64, b: pt.abi.Uint32, *, output: pt.abi.Uint64
    ) -> pt.Expr:
        return output.set(a.get() + b.get() + pt.Int(0xA190))

    @pt.ABIReturnSubroutine
    def fn_abi_annotations_0(
        a: pt.abi.Byte,
        b: pt.abi.StaticArray[pt.abi.Uint32, Literal[10]],
        c: pt.abi.DynamicArray[pt.abi.Bool],
    ) -> pt.Expr:
        return pt.Return()

    @pt.ABIReturnSubroutine
    def fn_abi_annotations_0_with_ret(
        a: pt.abi.Byte,
        b: pt.abi.StaticArray[pt.abi.Uint32, Literal[10]],
        c: pt.abi.DynamicArray[pt.abi.Bool],
        *,
        output: pt.abi.Byte,
    ):
        return output.set(a)

    @pt.ABIReturnSubroutine
    def fn_mixed_annotations_0(a: pt.ScratchVar, b: pt.Expr, c: pt.abi.Byte) -> pt.Expr:
        return pt.Seq(
            a.store(c.get() * pt.Int(0x0FF1CE) * b),
            pt.Return(),
        )

    @pt.ABIReturnSubroutine
    def fn_mixed_annotations_0_with_ret(
        a: pt.ScratchVar, b: pt.Expr, c: pt.abi.Byte, *, output: pt.abi.Uint64
    ) -> pt.Expr:
        return pt.Seq(
            a.store(c.get() * pt.Int(0x0FF1CE) * b),
            output.set(a.load()),
        )

    @pt.ABIReturnSubroutine
    def fn_mixed_annotation_1(
        a: pt.ScratchVar, b: pt.abi.StaticArray[pt.abi.Uint32, Literal[10]]
    ) -> pt.Expr:
        return pt.Seq(
            (intermediate := pt.abi.Uint32()).set(b[a.load() % pt.Int(10)]),
            a.store(intermediate.get()),
            pt.Return(),
        )

    @pt.ABIReturnSubroutine
    def fn_mixed_annotation_1_with_ret(
        a: pt.ScratchVar, b: pt.abi.Uint64, *, output: pt.abi.Bool
    ) -> pt.Expr:
        return output.set((a.load() + b.get()) % pt.Int(2))

    abi_u64 = pt.abi.Uint64()
    abi_u32 = pt.abi.Uint32()
    abi_byte = pt.abi.Byte()
    abi_static_u32_10 = pt.abi.StaticArray(
        pt.abi.StaticArrayTypeSpec(pt.abi.Uint32TypeSpec(), 10)
    )
    abi_dynamic_bool = pt.abi.DynamicArray(
        pt.abi.DynamicArrayTypeSpec(pt.abi.BoolTypeSpec())
    )
    sv = pt.ScratchVar()
    expr_int = pt.Int(1)

    cases = [
        ("vanilla 1", fn_log_add, [abi_u64, abi_u32], "void", None),
        (
            "vanilla 1 with wrong ABI type",
            fn_log_add,
            [abi_u64, abi_u64],
            None,
            pt.TealInputError,
        ),
        (
            "vanilla 1 with ABI return",
            fn_ret_add,
            [abi_u64, abi_u32],
            pt.abi.Uint64TypeSpec(),
            None,
        ),
        (
            "vanilla 1 with ABI return wrong typed",
            fn_ret_add,
            [abi_u32, abi_u64],
            None,
            pt.TealInputError,
        ),
        (
            "full ABI annotations no return",
            fn_abi_annotations_0,
            [abi_byte, abi_static_u32_10, abi_dynamic_bool],
            "void",
            None,
        ),
        (
            "full ABI annotations wrong input 0",
            fn_abi_annotations_0,
            [abi_u64, abi_static_u32_10, abi_dynamic_bool],
            None,
            pt.TealInputError,
        ),
        (
            "full ABI annotations with ABI return",
            fn_abi_annotations_0_with_ret,
            [abi_byte, abi_static_u32_10, abi_dynamic_bool],
            pt.abi.ByteTypeSpec(),
            None,
        ),
        (
            "full ABI annotations with ABI return wrong inputs",
            fn_abi_annotations_0_with_ret,
            [abi_byte, abi_dynamic_bool, abi_static_u32_10],
            None,
            pt.TealInputError,
        ),
        (
            "mixed with ABI annotations 0",
            fn_mixed_annotations_0,
            [sv, expr_int, abi_byte],
            "void",
            None,
        ),
        (
            "mixed with ABI annotations 0 wrong inputs",
            fn_mixed_annotations_0,
            [abi_u64, expr_int, abi_byte],
            None,
            pt.TealInputError,
        ),
        (
            "mixed with ABI annotations 0 with ABI return",
            fn_mixed_annotations_0_with_ret,
            [sv, expr_int, abi_byte],
            pt.abi.Uint64TypeSpec(),
            None,
        ),
        (
            "mixed with ABI annotations 0 with ABI return wrong inputs",
            fn_mixed_annotations_0_with_ret,
            [sv, expr_int, sv],
            None,
            pt.TealInputError,
        ),
        (
            "mixed with ABI annotations 1",
            fn_mixed_annotation_1,
            [sv, abi_static_u32_10],
            "void",
            None,
        ),
        (
            "mixed with ABI annotations 1 with ABI return",
            fn_mixed_annotation_1_with_ret,
            [sv, abi_u64],
            pt.abi.BoolTypeSpec(),
            None,
        ),
        (
            "mixed with ABI annotations 1 with ABI return wrong inputs",
            fn_mixed_annotation_1_with_ret,
            [expr_int, abi_static_u32_10],
            None,
            pt.TealInputError,
        ),
    ]

    for case_name, definition, args, ret_type, err in cases:
        assert definition.subroutine.argument_count() == len(args), case_name
        assert (
            definition.name() == definition.subroutine.implementation.__name__
        ), case_name

        if err is None:
            invocation = definition(*args)
            if ret_type == "void":
                assert isinstance(invocation, pt.SubroutineCall), case_name
                assert not invocation.has_return(), case_name
                assert invocation.args == args, case_name
            else:
                assert isinstance(invocation, pt.abi.ReturnedValue), case_name
                assert invocation.type_spec == ret_type
                assert isinstance(invocation.computation, pt.SubroutineCall), case_name
                assert not invocation.computation.has_return(), case_name
                assert invocation.computation.args == args, case_name
        else:
            try:
                with pytest.raises(err):
                    definition(*args)
            except Exception as e:
                assert (
                    not e
                ), f"EXPECTED ERROR of type {err}. encountered unexpected error during invocation case <{case_name}>: {e}"


def test_subroutine_definition_invalid():
    def fnWithDefaults(a, b=None):
        return pt.Return()

    def fnWithKeywordArgs(a, *, output):
        return pt.Return()

    def fnWithKeywordArgsWrongKWName(a, *, b: pt.abi.Uint64):
        return pt.Return()

    def fnWithMultipleABIKeywordArgs(a, *, b: pt.abi.Byte, c: pt.abi.Bool):
        return pt.Return()

    def fnWithVariableArgs(a, *b):
        return pt.Return()

    def fnWithNonExprReturnAnnotation(a, b) -> pt.TealType.uint64:
        return pt.Return()

    def fnWithNonExprParamAnnotation(a, b: pt.TealType.uint64):
        return pt.Return()

    def fnWithScratchVarSubclass(a, b: pt.DynamicScratchVar):
        return pt.Return()

    def fnReturningExprSubclass(a: pt.ScratchVar, b: pt.Expr) -> pt.Return:
        return pt.Return()

    def fnWithMixedAnns4AndBytesReturn(a: pt.Expr, b: pt.ScratchVar) -> pt.Bytes:
        return pt.Bytes("hello uwu")

    def fnWithMixedAnnsABIRet1(
        a: pt.Expr, b: pt.ScratchVar, c: pt.abi.Uint16
    ) -> pt.abi.StaticArray[pt.abi.Uint32, Literal[10]]:
        return pt.abi.StaticArray(
            pt.abi.StaticArrayTypeSpec(pt.abi.Uint32TypeSpec(), 10)
        )

    def fnWithMixedAnnsABIRet2(
        a: pt.Expr, b: pt.abi.Byte, c: pt.ScratchVar
    ) -> pt.abi.Uint64:
        return pt.abi.Uint64()

    cases = (
        (
            1,
            "TealInputError('Input to SubroutineDefinition is not callable'",
            "TealInputError('Input to ABIReturnSubroutine is not callable'",
        ),
        (
            None,
            "TealInputError('Input to SubroutineDefinition is not callable'",
            "TealInputError('Input to ABIReturnSubroutine is not callable'",
        ),
        (
            fnWithDefaults,
            "TealInputError('Function has a parameter with a default value, which is not allowed in a subroutine: b'",
            "TealInputError('Function has a parameter with a default value, which is not allowed in a subroutine: b'",
        ),
        (
            fnWithKeywordArgs,
            "TealInputError('Function has a parameter type that is not allowed in a subroutine: parameter output with type",
            "TealInputError('ABI return subroutine output-kwarg output must specify ABI type')",
        ),
        (
            fnWithKeywordArgsWrongKWName,
            "TealInputError('Function has a parameter type that is not allowed in a subroutine: parameter b with type",
            "TealInputError('ABI return subroutine output-kwarg name must be `output` at this moment",
        ),
        (
            fnWithMultipleABIKeywordArgs,
            "TealInputError('Function has a parameter type that is not allowed in a subroutine: parameter b with type",
            "multiple output arguments (2) with type annotations",
        ),
        (
            fnWithVariableArgs,
            "TealInputError('Function has a parameter type that is not allowed in a subroutine: parameter b with type",
            "Function has a parameter type that is not allowed in a subroutine: parameter b with type VAR_POSITIONAL",
        ),
        (
            fnWithNonExprReturnAnnotation,
            "Function has return of disallowed type TealType.uint64. Only Expr is allowed",
            "Function has return of disallowed type TealType.uint64. Only Expr is allowed",
        ),
        (
            fnWithNonExprParamAnnotation,
            "Function has parameter b of declared type TealType.uint64 which is not a class",
            "Function has parameter b of declared type TealType.uint64 which is not a class",
        ),
        (
            fnWithScratchVarSubclass,
            "Function has parameter b of disallowed type <class 'pyteal.DynamicScratchVar'>",
            "Function has parameter b of disallowed type <class 'pyteal.DynamicScratchVar'>",
        ),
        (
            fnReturningExprSubclass,
            "Function has return of disallowed type <class 'pyteal.Return'>",
            "Function has return of disallowed type <class 'pyteal.Return'>. Only Expr is allowed",
        ),
        (
            fnWithMixedAnns4AndBytesReturn,
            "Function has return of disallowed type <class 'pyteal.Bytes'>",
            "Function has return of disallowed type <class 'pyteal.Bytes'>. Only Expr is allowed",
        ),
        (
            fnWithMixedAnnsABIRet1,
            "Function has return of disallowed type pyteal.abi.StaticArray[pyteal.abi.Uint32, typing.Literal[10]]. "
            "Only Expr is allowed",
            "Function has return of disallowed type pyteal.abi.StaticArray[pyteal.abi.Uint32, typing.Literal[10]]. "
            "Only Expr is allowed",
        ),
        (
            fnWithMixedAnnsABIRet2,
            "Function has return of disallowed type <class 'pyteal.abi.Uint64'>. Only Expr is allowed",
            "Function has return of disallowed type <class 'pyteal.abi.Uint64'>. Only Expr is allowed",
        ),
    )

    for fn, sub_def_msg, abi_sub_def_msg in cases:
        with pytest.raises(pt.TealInputError) as e:
            print(f"case=[{sub_def_msg}]")
            pt.SubroutineDefinition(fn, pt.TealType.none)

        assert sub_def_msg in str(e), f"failed for case [{fn.__name__}]"

        with pytest.raises(pt.TealInputError) as e:
            print(f"case=[{abi_sub_def_msg}]")
            pt.ABIReturnSubroutine(fn)

        assert abi_sub_def_msg in str(e), f"failed for case[{fn.__name__}]"


def test_subroutine_declaration():
    cases = (
        (pt.TealType.none, pt.Return()),
        (pt.TealType.uint64, pt.Return(pt.Int(1))),
        (pt.TealType.uint64, pt.Int(1)),
        (pt.TealType.bytes, pt.Bytes("value")),
        (pt.TealType.anytype, pt.App.globalGet(pt.Bytes("key"))),
    )

    for (returnType, value) in cases:

        def mySubroutine():
            return value

        definition = pt.SubroutineDefinition(mySubroutine, returnType)

        declaration = pt.SubroutineDeclaration(definition, value)
        assert declaration.type_of() == value.type_of()
        assert declaration.has_return() == value.has_return()

        options.currentSubroutine = definition
        assert declaration.__teal__(options) == value.__teal__(options)
        options.setSubroutine(None)


def test_subroutine_call():
    def mySubroutine():
        return pt.Return()

    returnTypes = (
        pt.TealType.uint64,
        pt.TealType.bytes,
        pt.TealType.anytype,
        pt.TealType.none,
    )

    argCases = (
        [],
        [pt.Int(1)],
        [pt.Int(1), pt.Bytes("value")],
    )

    for returnType in returnTypes:
        definition = pt.SubroutineDefinition(mySubroutine, returnType)

        for args in argCases:
            expr = pt.SubroutineCall(definition, args)

            assert expr.type_of() == returnType
            assert not expr.has_return()

            expected, _ = pt.TealBlock.FromOp(
                options, pt.TealOp(expr, pt.Op.callsub, definition), *args
            )

            actual, _ = expr.__teal__(options)

            assert actual == expected


def test_decorator():
    assert callable(pt.Subroutine)
    assert callable(pt.Subroutine(pt.TealType.anytype))

    @pt.Subroutine(pt.TealType.none)
    def mySubroutine(a):
        return pt.Return()

    assert isinstance(mySubroutine, pt.SubroutineFnWrapper)

    invocation = mySubroutine(pt.Int(1))
    assert isinstance(invocation, pt.SubroutineCall)

    with pytest.raises(pt.TealInputError):
        mySubroutine()

    with pytest.raises(pt.TealInputError):
        mySubroutine(pt.Int(1), pt.Int(2))

    with pytest.raises(pt.TealInputError):
        mySubroutine(pt.Pop(pt.Int(1)))

    with pytest.raises(pt.TealInputError):
        mySubroutine(1)

    with pytest.raises(pt.TealInputError):
        mySubroutine(a=pt.Int(1))


def test_evaluate_subroutine_no_args():
    cases = (
        (pt.TealType.none, pt.Return()),
        (pt.TealType.uint64, pt.Int(1) + pt.Int(2)),
        (pt.TealType.uint64, pt.Return(pt.Int(1) + pt.Int(2))),
        (pt.TealType.bytes, pt.Bytes("value")),
        (pt.TealType.bytes, pt.Return(pt.Bytes("value"))),
    )

    for (returnType, returnValue) in cases:

        def mySubroutine():
            return returnValue

        definition = pt.SubroutineDefinition(mySubroutine, returnType)
        declaration = evaluate_subroutine(definition)

        assert isinstance(declaration, pt.SubroutineDeclaration)
        assert declaration.subroutine is definition

        assert declaration.type_of() == returnValue.type_of()
        assert declaration.has_return() == returnValue.has_return()

        options.setSubroutine(definition)
        expected, _ = pt.Seq([returnValue]).__teal__(options)

        actual, _ = declaration.__teal__(options)
        options.setSubroutine(None)
        assert actual == expected


def test_evaluate_subroutine_1_arg():
    cases = (
        (pt.TealType.none, pt.Return()),
        (pt.TealType.uint64, pt.Int(1) + pt.Int(2)),
        (pt.TealType.uint64, pt.Return(pt.Int(1) + pt.Int(2))),
        (pt.TealType.bytes, pt.Bytes("value")),
        (pt.TealType.bytes, pt.Return(pt.Bytes("value"))),
    )

    for (returnType, returnValue) in cases:
        argSlots: List[pt.ScratchSlot] = []

        def mySubroutine(a1):
            assert isinstance(a1, pt.ScratchLoad)
            argSlots.append(a1.slot)
            return returnValue

        definition = pt.SubroutineDefinition(mySubroutine, returnType)
        declaration = evaluate_subroutine(definition)

        assert isinstance(declaration, pt.SubroutineDeclaration)
        assert declaration.subroutine is definition

        assert declaration.type_of() == returnValue.type_of()
        assert declaration.has_return() == returnValue.has_return()

        assert isinstance(declaration.body, pt.Seq)
        assert len(declaration.body.args) == 2

        assert isinstance(declaration.body.args[0], pt.ScratchStackStore)

        assert declaration.body.args[0].slot is argSlots[-1]

        options.setSubroutine(definition)
        expected, _ = pt.Seq([declaration.body.args[0], returnValue]).__teal__(options)

        actual, _ = declaration.__teal__(options)
        options.setSubroutine(None)
        assert actual == expected


def test_evaluate_subroutine_2_args():
    cases = (
        (pt.TealType.none, pt.Return()),
        (pt.TealType.uint64, pt.Int(1) + pt.Int(2)),
        (pt.TealType.uint64, pt.Return(pt.Int(1) + pt.Int(2))),
        (pt.TealType.bytes, pt.Bytes("value")),
        (pt.TealType.bytes, pt.Return(pt.Bytes("value"))),
    )

    for (returnType, returnValue) in cases:
        argSlots: List[pt.ScratchSlot] = []

        def mySubroutine(a1, a2):
            assert isinstance(a1, pt.ScratchLoad)
            argSlots.append(a1.slot)
            assert isinstance(a2, pt.ScratchLoad)
            argSlots.append(a2.slot)
            return returnValue

        definition = pt.SubroutineDefinition(mySubroutine, returnType)

        declaration = evaluate_subroutine(definition)

        assert isinstance(declaration, pt.SubroutineDeclaration)
        assert declaration.subroutine is definition

        assert declaration.type_of() == returnValue.type_of()
        assert declaration.has_return() == returnValue.has_return()

        assert isinstance(declaration.body, pt.Seq)
        assert len(declaration.body.args) == 3

        assert isinstance(declaration.body.args[0], pt.ScratchStackStore)
        assert isinstance(declaration.body.args[1], pt.ScratchStackStore)

        assert declaration.body.args[0].slot is argSlots[-1]
        assert declaration.body.args[1].slot is argSlots[-2]

        options.setSubroutine(definition)
        expected, _ = pt.Seq(
            [declaration.body.args[0], declaration.body.args[1], returnValue]
        ).__teal__(options)

        actual, _ = declaration.__teal__(options)
        options.setSubroutine(None)
        assert actual == expected


def test_evaluate_subroutine_10_args():
    cases = (
        (pt.TealType.none, pt.Return()),
        (pt.TealType.uint64, pt.Int(1) + pt.Int(2)),
        (pt.TealType.uint64, pt.Return(pt.Int(1) + pt.Int(2))),
        (pt.TealType.bytes, pt.Bytes("value")),
        (pt.TealType.bytes, pt.Return(pt.Bytes("value"))),
    )

    for (returnType, returnValue) in cases:
        argSlots: List[pt.ScratchSlot] = []

        def mySubroutine(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10):
            for a in (a1, a2, a3, a4, a5, a6, a7, a8, a9, a10):
                assert isinstance(a, pt.ScratchLoad)
                argSlots.append(a.slot)
            return returnValue

        definition = pt.SubroutineDefinition(mySubroutine, returnType)
        declaration = evaluate_subroutine(definition)

        assert isinstance(declaration, pt.SubroutineDeclaration)
        assert declaration.subroutine is definition

        assert declaration.type_of() == returnValue.type_of()
        assert declaration.has_return() == returnValue.has_return()

        assert isinstance(declaration.body, pt.Seq)
        assert len(declaration.body.args) == 11

        for i in range(10):
            assert isinstance(declaration.body.args[i], pt.ScratchStackStore)

        for i in range(10):
            assert declaration.body.args[i].slot is argSlots[-i - 1]

        options.setSubroutine(definition)
        expected, _ = pt.Seq(declaration.body.args[:10] + [returnValue]).__teal__(
            options
        )

        actual, _ = declaration.__teal__(options)
        options.setSubroutine(None)
        assert actual == expected


def test_docstring_parsing_with_different_format():
    short_desc = "Example of a ABIReturnSubroutine with short description docstring."
    a_doc = "an abi Uint64 value"
    return_doc = "A PyTeal expression that sets output Uint64 value as argument a."
    long_desc = """Example first line.

    This is a second line.

    This is a third line that's so long it has to wrap in order to fit properly
    in a line of source code.
    """
    expected_long_desc = "Example first line.\nThis is a second line.\nThis is a third line that's so long it has to wrap in order to fit properly in a line of source code."

    def documented_method(a: pt.abi.Uint64, *, output: pt.abi.Uint64):
        return output.set(a)

    # Google format
    documented_method.__doc__ = f"""{short_desc}

    Args:
        a: {a_doc}

    Returns:
        {return_doc}
    """

    mspec_dict = pt.ABIReturnSubroutine(documented_method).method_spec().dictify()
    assert mspec_dict["desc"] == short_desc
    assert mspec_dict["args"][0]["desc"] == a_doc
    assert mspec_dict["returns"]["desc"] == return_doc

    # epy format
    documented_method.__doc__ = f"""
    {short_desc}

    @param a: {a_doc}
    @return: {return_doc}
    """

    mspec_dict = ABIReturnSubroutine(documented_method).method_spec().dictify()
    assert mspec_dict["desc"] == short_desc
    assert mspec_dict["args"][0]["desc"] == a_doc
    assert mspec_dict["returns"]["desc"] == return_doc

    # numpy format
    documented_method.__doc__ = f"""{short_desc}

    Parameters
    ----------
    a:
        an abi Uint64 value
    output:
        {a_doc}

    Returns
    -------
    uint64
        {return_doc}
    """

    mspec_dict = ABIReturnSubroutine(documented_method).method_spec().dictify()
    assert mspec_dict["desc"] == short_desc
    assert mspec_dict["args"][0]["desc"] == a_doc
    assert mspec_dict["returns"]["desc"] == return_doc

    # rst format
    documented_method.__doc__ = f"""{short_desc}

    :param a: {a_doc}
    :returns: {return_doc}
    """

    mspec_dict = ABIReturnSubroutine(documented_method).method_spec().dictify()
    assert mspec_dict["desc"] == short_desc
    assert mspec_dict["args"][0]["desc"] == a_doc
    assert mspec_dict["returns"]["desc"] == return_doc

    # Short and long descriptions
    documented_method.__doc__ = f"""{long_desc}

    :param a: {a_doc}
    :returns: {return_doc}
    """

    mspec_dict = ABIReturnSubroutine(documented_method).method_spec().dictify()
    assert mspec_dict["desc"] == expected_long_desc
    assert mspec_dict["args"][0]["desc"] == a_doc
    assert mspec_dict["returns"]["desc"] == return_doc

    # Only long description
    # Short description is defined as being on the first line, so by introducing
    # long_desc on the second, there is no short description.
    documented_method.__doc__ = f"""
    {long_desc}

    :param a: {a_doc}
    :returns: {return_doc}
    """

    mspec_dict = ABIReturnSubroutine(documented_method).method_spec().dictify()
    assert mspec_dict["desc"] == expected_long_desc
    assert mspec_dict["args"][0]["desc"] == a_doc
    assert mspec_dict["returns"]["desc"] == return_doc

    # No description
    documented_method.__doc__ = f"""

    :param a: {a_doc}
    :returns: {return_doc}
    """
    mspec_dict = ABIReturnSubroutine(documented_method).method_spec().dictify()
    assert "desc" not in mspec_dict
    assert mspec_dict["args"][0]["desc"] == a_doc
    assert mspec_dict["returns"]["desc"] == return_doc

    # No doc
    documented_method.__doc__ = None
    mspec_dict = ABIReturnSubroutine(documented_method).method_spec().dictify()
    assert "desc" not in mspec_dict
    assert len(mspec_dict["args"]) == 1

    algobank_example = """Withdraw an amount of Algos held by this app.

    The sender of this method call will be the source of the Algos, and the destination will be
    the `recipient` argument.

    The Algos will be transferred to the recipient using an inner transaction whose fee is set
    to 0, meaning the caller's transaction must include a surplus fee to cover the inner
    transaction.

    Args:
        amount: The amount of Algos requested to be withdraw, in microAlgos. This method will fail
            if this amount exceeds the amount of Algos held by this app for the method call sender.
        recipient: An account who will receive the withdrawn Algos. This may or may not be the same
            as the method call sender.
    """

    # algobank example
    def withdraw(amount: pt.abi.Uint64, recipient: pt.abi.Account):
        return pt.Assert(pt.Int(1))

    withdraw.__doc__ = algobank_example

    mspec_dict = ABIReturnSubroutine(withdraw).method_spec().dictify()
    assert (
        mspec_dict["desc"]
        == "Withdraw an amount of Algos held by this app.\nThe sender of this method call will be the source of the Algos, "
        + "and the destination will be the `recipient` argument.\nThe Algos will be transferred to the recipient using an inner transaction whose fee is "
        + "set to 0, meaning the caller's transaction must include a surplus fee to cover the inner transaction."
    )
    assert (
        mspec_dict["args"][0]["desc"]
        == "The amount of Algos requested to be withdraw, in microAlgos. This method will fail if this amount exceeds "
        + "the amount of Algos held by this app for the method call sender."
    )
    assert (
        mspec_dict["args"][1]["desc"]
        == "An account who will receive the withdrawn Algos. This may or may not be the same as the method call sender."
    )
    assert "desc" not in mspec_dict["returns"]


def test_relaxed_abi_arg():
    @pt.Subroutine(pt.TealType.none)
    def txn_checker(t: pt.abi.Transaction):
        return pt.Seq(
            pt.Assert(
                pt.Or(
                    t.get().type_enum() == pt.TxnType.Payment,
                    t.get().type_enum() == pt.TxnType.AssetTransfer,
                )
            )
        )

    # Pass a payment transaction to the subroutine that expects a Transaction type
    # this will fail if we don't have a relaxed type check for the transaction type
    program = pt.Seq(
        (p := pt.abi.PaymentTransaction())._set_index(pt.Txn.group_index() - pt.Int(1)),
        txn_checker(p),
        pt.Int(1),
    )

    pt.compileTeal(program, mode=pt.Mode.Application, version=7)


def test_override_abi_method_name():
    def abi_meth(a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64):
        return output.set(a.get() + b.get())

    mspec = ABIReturnSubroutine(abi_meth).method_spec().dictify()
    assert mspec["name"] == "abi_meth"

    mspec = ABIReturnSubroutine(abi_meth, overriding_name="add").method_spec().dictify()
    assert mspec["name"] == "add"

    @ABIReturnSubroutine.name_override("overriden_add")
    def abi_meth_2(a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64):
        return output.set(a.get() + b.get())

    mspec = abi_meth_2.method_spec().dictify()
    assert mspec["name"] == "overriden_add"
